import logging
from pathlib import Path
from typing import Union, overload, Any
from collections.abc import Iterable, Sequence

from transformers import BatchEncoding, PreTrainedTokenizerBase
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy

from kazu.data.data import Document, Entity, Section

logger = logging.getLogger(__name__)


def find_document_from_entity(docs: list[Document], entity: Entity) -> Document:
    """For a given entity and a list of docs, find the doc the entity belongs
    to.

    :param docs:
    :param entity:
    :return:
    """
    for doc in docs:
        if entity in doc.get_entities():
            return doc
    raise RuntimeError(f"Error! Entity {entity} is not attached to a document")


def documents_to_id_section_map(docs: list[Document]) -> dict[int, Section]:
    """Return a map of documents, indexed by order of sections.

    :param docs:
    :return:
    """
    result = {}
    i = 0
    for doc in docs:
        for section in doc.sections:
            result[i] = section
            i += 1
    return result


def documents_to_document_section_batch_encodings_map(
    docs: list[Document],
    tokenizer: PreTrainedTokenizerBase,
    stride: int = 128,
    max_length: int = 512,
) -> tuple[BatchEncoding, dict[int, Section]]:
    """Convert documents into a BatchEncoding. Also returns a list of <int +
    section> for the resulting encoding.

    :param docs:
    :param tokenizer:
    :param stride:
    :param max_length:
    :return:
    """
    id_section_map = documents_to_id_section_map(docs)
    strings = [section.text for section in id_section_map.values()]

    batch_encodings = tokenizer(
        strings,
        stride=stride,
        max_length=max_length,
        truncation=TruncationStrategy.LONGEST_FIRST,
        return_overflowing_tokens=True,
        padding=PaddingStrategy.MAX_LENGTH,
    )
    return batch_encodings, id_section_map


def get_match_entity_class_hash(ent: Entity) -> int:
    return hash(
        (
            ent.match,
            ent.entity_class,
        )
    )


PathLike = Union[str, Path]


def as_path(p: PathLike) -> Path:
    return p if isinstance(p, Path) else Path(p)


class EntityClassFilter:
    """A condition that returns True if a document has any entities that match
    the class of the required_entity_classes."""

    def __init__(self, required_entity_classes: Iterable[str]):
        """

        :param required_entity_classes: list of str, specifying entity classes to assess
        """
        self.required_entities = set(required_entity_classes)

    def __call__(self, document: Document) -> bool:
        return any(
            (entity.entity_class in self.required_entities for entity in document.get_entities())
        )


@overload
def _create_ngrams_iter(tokens: str, n: int) -> Iterable[str]:
    pass


@overload
def _create_ngrams_iter(tokens: Sequence[str], n: int) -> Iterable[Sequence[str]]:
    pass


def _create_ngrams_iter(tokens, n=2):
    """Yields ngrams of the input as a sequence of strings.

    Tokens can be a single string where each token is a character, or an
    Iterable of words. If tokens is a str
    """
    num_tokens = len(tokens)
    for i in range(num_tokens):
        ngram_end_index = i + n
        if ngram_end_index > num_tokens:
            # ngram would extend beyond end of parts
            # it's ok for it to be the same number though as otherwise
            # we don't get the final word of ngrams at the right of parts
            break
        yield tokens[i : i + n]


def create_char_ngrams(s: str, n: int = 2) -> list[str]:
    """Return list of char ngrams as a string."""
    return list(_create_ngrams_iter(s, n))


def create_word_ngrams(s: str, n: int = 2) -> list[str]:
    """Return list of word ngrams as a single space-separated string."""
    words = s.split(" ")
    ngrams_iter = _create_ngrams_iter(words, n)
    return [" ".join(ngram) for ngram in ngrams_iter]


class Singleton(type):
    _instances: dict[type, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @staticmethod
    def clear_all():
        logger.warning(
            "When clearing singletons, check that instances of classes with metaclass=Singleton are "
            "not used as class fields, as this will cause unexpected behaviour. Also note that any existing "
            "classes that use a Singleton will need to be re-instantiated to work correctly!. Much pain "
            "can be avoided by observing this simple principle! Some people consider Singletons an "
            "anti-pattern, and this is a reason why!!!"
        )
        Singleton._instances.clear()



from collections.abc import Iterable

from spacy.tokens import Doc, Span, Token

from kazu.data.data import Section, Entity
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.utils.spacy_pipeline import SpacyPipelines, BASIC_PIPELINE_NAME, basic_spacy_pipeline


class SpacyToKazuObjectMapper:
    """Maps entities and text from a :class:`.Section` to a Spacy `Doc
    <https://spacy.io/api/doc>`_ using :func:`.basic_spacy_pipeline`\\."""

    def __init__(self, parsers: Iterable[OntologyParser]):
        """

        :param parsers: parsers are required so that all entity classes can be set as spacy extensions
        """
        for parser in parsers:
            Token.set_extension(parser.entity_class, default=False, force=True)
        self.spacy_pipelines = SpacyPipelines()
        self.spacy_pipelines.add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)

    def __call__(self, section: Section) -> dict[Entity, Span]:
        """Convert a :class:`.Section` into a dictionary of :class:`.Entity` to
        Spacy `Span <https://spacy.io/api/span>`_ mappings."""

        spacy_doc: Doc = self.spacy_pipelines.process_single(section.text, BASIC_PIPELINE_NAME)
        ent_to_span = {}
        for entity in section.entities:
            entity_class = entity.entity_class
            span = spacy_doc.char_span(
                start_idx=entity.start,
                end_idx=entity.end,
                label=entity_class,
                alignment_mode="expand",
            )
            if span is not None:
                ent_to_span[entity] = span
                for token in span:
                    token._.set(entity_class, True)
        return ent_to_span



import logging
from collections import defaultdict
from collections.abc import Iterable

import ahocorasick

from kazu.data.data import Document, Entity, SynonymTermWithMetrics, MentionConfidence
from kazu.database.in_memory_db import SynonymDatabase, ParserName, NormalisedSynonymStr
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.steps import document_iterating_step
from kazu.steps.step import ParserDependentStep
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.curated_term_tools import filter_curations_for_ner
from kazu.utils.spacy_pipeline import SpacyPipelines, basic_spacy_pipeline, BASIC_PIPELINE_NAME

logger = logging.getLogger(__name__)

EntityClass = str
CaseSensitive = bool
EntityInfoToOntologyInfoMapping = defaultdict[
    tuple[EntityClass, MentionConfidence, CaseSensitive, NormalisedSynonymStr], set[ParserName]
]


class MemoryEfficientStringMatchingStep(ParserDependentStep):
    """A wrapper for the ahocorasick algorithm.

    In testing, this implementation is comparable in speed to a Spacy
    `PhraseMatcher <https://spacy.io/api/phrasematcher>`_\\,
    and uses a fraction of the memory. Since this implementation is unaware
    of NLP concepts such as tokenization, we backfill this capability by
    checking for word boundaries with a custom spacy tokenizer.
    """

    def __init__(self, parsers: Iterable[OntologyParser]):
        super().__init__(parsers)
        self.parsers = parsers
        self.automaton = self._create_automaton()
        self.spacy_pipelines = SpacyPipelines()
        self.spacy_pipelines.add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)
        self.synonym_db = SynonymDatabase()

    @kazu_disk_cache.memoize(ignore={0})
    def _create_automaton(self) -> ahocorasick.Automaton:
        """Create `ahocorasick.Automaton
        <https://pyahocorasick.readthedocs.io/en/latest/#automaton-class>`_\\ s
        for parsers."""
        key_to_ontology_info: defaultdict[str, EntityInfoToOntologyInfoMapping] = defaultdict(
            lambda: defaultdict(set)
        )

        logger.info("Creating ahocorasick Automaton")
        for parser in self.parsers:
            parser_curations = parser.populate_databases(return_curations=True)
            if parser_curations is None:
                logger.warning(
                    "tried to create ahocorasick data for parser %s, but no curations are available",
                    parser.name,
                )
                continue
            if len(parser_curations) == 0:
                logger.warning(
                    "tried to create ahocorasick data for parser %s, but no curations were produced",
                    parser.name,
                )
                continue

            for curation in filter_curations_for_ner(parser_curations, parser):
                # a curation can have different term_norms for different parsers,
                # since the string normalizer's output depends on the entity class.
                # Also, a curation may exist in multiple SynonymTerm.terms
                term_norm = curation.term_norm_for_linking(parser.entity_class)
                entity_key = (
                    parser.entity_class,
                    curation.mention_confidence,
                    curation.case_sensitive,
                    term_norm,
                )
                key_to_ontology_info[curation.curated_synonym][entity_key].add(parser.name)

        if len(key_to_ontology_info) > 0:
            case_insensitive_automaton = ahocorasick.Automaton()
            for key, entity_to_ontology_info in key_to_ontology_info.items():
                case_insensitive_automaton.add_word(key.lower(), (key, entity_to_ontology_info))
            case_insensitive_automaton.make_automaton()
        else:
            raise RuntimeError(
                f"No valid curations were available for {self.__class__.__name__}. Either remove this step "
                "from your pipeline or ensure at least one valid curation is generated from the chosen "
                "parsers."
            )

        return case_insensitive_automaton

    def _word_is_valid(
        self, start_char: int, end_char: int, starts: set[int], ends: set[int]
    ) -> bool:
        return start_char in starts and end_char in ends

    def _case_matches(self, actual_match: str, original_casing: str, case_sensitive: bool) -> bool:
        # no need to check the strings actually match if it's not case sensitive
        # as this is already done as part of the string matching, so we know that
        # actual_match.lower() == original_casing.lower()
        return (not case_sensitive) or actual_match == original_casing

    def _process_automaton(
        self,
        automaton: ahocorasick.Automaton,
        matchable_text: str,
        original_text: str,
        starts: set[int],
        ends: set[int],
    ) -> list[Entity]:
        entities = []
        for end_index, (match_key, ontology_dict) in automaton.iter(matchable_text):
            start_index = end_index - len(match_key) + 1
            original_match = original_text[start_index : end_index + 1]
            if self._word_is_valid(start_index, end_index, starts, ends):
                for entity_info, parser_name_set in ontology_dict.items():
                    (
                        entity_class,
                        confidence,
                        case_sensitive,
                        term_norm,
                    ) = entity_info
                    # filter out cases where we've specified we only want exact matches, but we only have a lowercase match
                    if not self._case_matches(
                        actual_match=original_match,
                        original_casing=match_key,
                        case_sensitive=case_sensitive,
                    ):
                        continue
                    terms = [
                        SynonymTermWithMetrics.from_synonym_term(
                            self.synonym_db.get(parser_name, term_norm), exact_match=True
                        )
                        for parser_name in parser_name_set
                    ]
                    e = Entity.load_contiguous_entity(
                        start=start_index,
                        end=end_index + 1,
                        match=original_match,
                        entity_class=entity_class,
                        namespace=self.namespace(),
                        mention_confidence=confidence,
                    )
                    e.update_terms(terms)
                    entities.append(e)
        return entities

    @document_iterating_step
    def __call__(self, doc: Document) -> None:

        for section in doc.sections:
            spacy_doc = self.spacy_pipelines.process_single(section.text, BASIC_PIPELINE_NAME)
            starts, ends = set(), set()
            for tok in spacy_doc:
                starts.add(tok.idx)
                ends.add(tok.idx + len(tok) - 1)

            section.entities.extend(
                self._process_automaton(
                    self.automaton, section.text.lower(), section.text, starts, ends
                )
            )



import os
from typing import Optional
from collections.abc import Callable

try:
    from py4j.java_gateway import JavaGateway
except ImportError as e:
    raise ImportError(
        "To use SethStep, you need to install py4j.\n"
        "You can either install py4j yourself, or install kazu[all-steps].\n"
    ) from e

from kazu.data.data import Document, Entity, Mapping, StringMatchConfidence
from kazu.steps import Step, document_iterating_step


SETH_METADATA_KEY = "seth"


class SethStep(Step):
    """A Step that calls SETH (SNP Extraction Tool for Human Variations) over
    py4j.

    .. attention::

        To use this step, you will need `py4j <https://www.py4j.org>`_ installed,
        which is not installed as part of the default kazu install because this
        step isn't used as part of the default pipeline.

        You can either do:

        .. code-block:: console

            $ pip install py4j

        Or you can install required dependencies for all steps included in kazu
        with:

        .. code-block:: console

            $ pip install kazu[all-steps]

    Paper:

    | Thomas, P., Rocktäschel, T., Hakenberg, J., Mayer, L., and Leser, U. (2016).
    | `SETH detects and normalizes genetic variants in text. <https://pubmed.ncbi.nlm.nih.gov/27256315/>`_
    | Bioinformatics (2016)

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @Article{SETH2016,
        Title= {SETH detects and normalizes genetic variants in text.},
        Author= {Thomas, Philippe and Rockt{\"{a}}schel, Tim and Hakenberg, J{\"{o}}rg and Lichtblau, Yvonne and Leser, Ulf},
        Journal= {Bioinformatics},
        Year= {2016},
        Month= {Jun},
        Doi= {10.1093/bioinformatics/btw234},
        Language = {eng},
        Medline-pst = {aheadofprint},
        Pmid = {27256315},
        Url = {http://dx.doi.org/10.1093/bioinformatics/btw234}
        }

    .. raw:: html

        </details>
    """

    def __init__(
        self,
        entity_class: str,
        seth_fatjar_path: str,
        java_home: str,
        condition: Optional[Callable[[Document], bool]] = None,
    ):
        """
        :param entity_class: the entity_class to assign to any Entities that emerge
        :param seth_fatjar_path: path to a py4j fatjar, containing SETH dependencies
        :param java_home: path to installed java runtime
        :param condition: Since SETH can be slow, we can optionally specify a callable, so that
            any documents that don't contain pre-existing gene/protein entities are not processed
        """
        self.condition = condition
        if not os.path.exists(seth_fatjar_path):
            raise RuntimeError(f"required jar: {seth_fatjar_path} not found")
        self.gateway = JavaGateway.launch_gateway(
            classpath=seth_fatjar_path,
            die_on_exit=True,
            java_path=os.path.join(java_home, "bin", "java"),
        )
        self.seth = self.gateway.jvm.com.astrazeneca.kazu.SethRunner()
        self.entity_class = entity_class

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        if self.condition and not self.condition(doc):
            # skip this document
            return

        for section in doc.sections:
            mutation_lst = self.seth.findMutations(section.text)
            entities = []
            for java_mutation_dict in mutation_lst:
                python_dict = dict(java_mutation_dict)
                entities.append(
                    Entity.from_spans(
                        spans=[
                            (
                                python_dict.pop("start"),
                                python_dict.pop("end"),
                            )
                        ],
                        text=section.text,
                        entity_class=self.entity_class,
                        namespace=self.namespace(),
                        metadata={SETH_METADATA_KEY: python_dict},
                        mappings=[
                            Mapping(
                                default_label=self.entity_class,
                                source=self.entity_class,
                                parser_name="n/a",
                                idx=self.entity_class,
                                string_match_strategy=self.namespace(),
                                disambiguation_strategy=None,
                                string_match_confidence=StringMatchConfidence.PROBABLE,
                            )
                        ],
                    ),
                )
            section.entities.extend(entities)



"""Original source:

https://raw.githubusercontent.com/amitripshtos/starlette-jwt/master/starlette_jwt/middleware.py


All code in this file is originally licensed under the BSD 3-Clause License as part of starlette-jwt above.

Copyright (c) 2018, Amit Ripshtos All rights reserved.

.. raw:: html

    <details>
    <summary>Full License</summary>

Copyright (c) 2018, Amit Ripshtos
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

.. raw:: html

    </details>
"""

import logging
import uuid
from typing import Optional, Union

import jwt
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
)
from starlette.requests import Request, HTTPConnection
from starlette.responses import JSONResponse

from kazu.web.routes import NO_AUTH_ENDPOINTS, NO_AUTH_DIRS

logger = logging.getLogger("ray")


class JWTUser(BaseUser):
    def __init__(self, username: str, token: str, payload: dict) -> None:
        self.username = username
        self.token = token
        self.payload = payload

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.username


class JWTAuthenticationBackend(AuthenticationBackend):
    def __init__(
        self,
        secret_key: str,
        algorithm: str = "HS256",
        prefix: str = "Bearer",
        username_field: str = "username",
        audience: Optional[str] = None,
        options: Optional[dict] = None,
    ) -> None:
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.prefix = prefix
        self.username_field = username_field
        self.audience = audience
        self.options = options or dict()

    @classmethod
    def get_token_from_header(cls, authorization: str, prefix: str) -> str:
        """Parses the Authorization header and returns only the token."""
        try:
            scheme, token = authorization.split()
        except ValueError:
            raise AuthenticationError("Could not separate Authorization scheme and token")
        if scheme.lower() != prefix.lower():
            raise AuthenticationError(f"Authorization scheme {scheme} is not supported")
        return token

    async def authenticate(
        self, request: HTTPConnection
    ) -> Union[None, tuple[AuthCredentials, BaseUser]]:
        req_id = str(uuid.uuid4())
        request.headers.__dict__["_list"].append(
            (
                "X-request-id".encode(),
                req_id.encode(),
            )
        )
        path: str = request.scope["raw_path"].decode()
        if path in NO_AUTH_ENDPOINTS or any(path.startswith(x) for x in NO_AUTH_DIRS):
            logger.info("ID: %s Request to %s, no authentication required", req_id, path)
            return None

        if "Authorization" not in request.headers:
            message = (
                f"ID: {req_id} No 'Authorization' header specified: please use a valid Bearer token"
            )
            logger.warn(message)
            raise AuthenticationError(message)
        auth = request.headers["Authorization"]
        token = self.get_token_from_header(authorization=auth, prefix=self.prefix)
        try:
            payload = jwt.decode(
                token,
                key=self.secret_key,
                algorithms=[self.algorithm],
                audience=self.audience,
                options=self.options,
            )
        except jwt.InvalidTokenError as e:
            logger.warn("ID: %s %s", req_id, e)
            raise AuthenticationError(f"ID: {req_id} {e}")
        username = payload[self.username_field]
        logger.info("ID: %s Received request from %s", req_id, username)
        return AuthCredentials(["authenticated"]), JWTUser(
            username=username, token=token, payload=payload
        )


def on_auth_error(request: Request, exc: Exception) -> JSONResponse:
    return JSONResponse({"error": str(exc)}, status_code=401)



import logging
from functools import partial
from typing import Optional, cast
from collections.abc import Iterable, Callable

import torch
from pytorch_lightning import Trainer
from torch import Tensor, sigmoid, softmax
from torch.utils.data import DataLoader
from transformers import (
    AutoModelForTokenClassification,
    AutoConfig,
    AutoTokenizer,
    DataCollatorWithPadding,
    BatchEncoding,
    PreTrainedTokenizerBase,
)
from transformers.file_utils import PaddingStrategy

from kazu.data.data import Document, Section
from kazu.data.pytorch import HFDataset
from kazu.steps import Step, document_batch_step
from kazu.steps.ner.hf_lightning_wrappers import PLAutoModelForTokenClassification
from kazu.steps.ner.tokenized_word_processor import TokenizedWordProcessor, TokenizedWord
from kazu.steps.ner.entity_post_processing import NonContiguousEntitySplitter
from kazu.utils.utils import documents_to_document_section_batch_encodings_map

logger = logging.getLogger(__name__)


class TransformersModelForTokenClassificationNerStep(Step):
    """
    A wrapper for
    :class:`transformers.AutoModelForTokenClassification`\\ . This implementation uses a sliding
    window concept to process large documents that don't fit into the maximum sequence length
    allowed by a model. Resulting token labels are then post processed by
    :class:`~kazu.steps.ner.tokenized_word_processor.TokenizedWordProcessor`.
    """

    def __init__(
        self,
        path: str,
        batch_size: int,
        stride: int,
        max_sequence_length: int,
        trainer: Trainer,
        labels: list[str],
        detect_subspans: bool = False,
        threshold: Optional[float] = None,
        entity_splitter: Optional[NonContiguousEntitySplitter] = None,
        strip_re: Optional[dict[str, str]] = None,
    ):
        """

        :param path: path to HF model, config and tokenizer. Passed to HF .from_pretrained()
        :param batch_size: batch size for dataloader
        :param stride: passed to HF tokenizers (for splitting long docs)
        :param max_sequence_length: passed to HF tokenizers (for splitting long docs)
        :param trainer:
        :param detect_subspans: attempt to detect nested entities (threshold must be configured)
        :param threshold: the confidence threshold used to detect nested entities
        :param entity_splitter: to detect non-contiguous entities if provided
        :param strip_re: passed to :class:`~kazu.steps.ner.tokenized_word_processor.TokenizedWordProcessor`
        """
        self.entity_splitter = entity_splitter
        if max_sequence_length % 2 != 0:
            raise RuntimeError(
                "max_sequence_length must %2 ==0 in order for correct document windowing"
            )
        self.max_sequence_length = max_sequence_length
        if stride % 2 != 0:
            raise RuntimeError("stride must %2 ==0 in order for correct document windowing")
        self.stride = stride
        self.batch_size = batch_size
        self.config = AutoConfig.from_pretrained(path)
        self.tokeniser: PreTrainedTokenizerBase = AutoTokenizer.from_pretrained(
            path, config=self.config
        )
        self.model = AutoModelForTokenClassification.from_pretrained(path, config=self.config)
        self.model = PLAutoModelForTokenClassification(self.model).eval()
        self.trainer = trainer
        self.activation_fn = cast(
            Callable[[Tensor], Tensor], sigmoid if detect_subspans else partial(softmax, dim=-1)
        )
        self.tokenized_word_processor = TokenizedWordProcessor(
            detect_subspans=detect_subspans,
            confidence_threshold=threshold,
            id2label=self.id2labels_from_label_list(labels),
            strip_re=strip_re,
        )

    @document_batch_step
    def __call__(self, docs: list[Document]) -> None:
        loader, id_section_map = self.get_dataloader(docs)
        # need this so mypy knows to expect the dataset to have encodings
        dataset = cast(HFDataset, loader.dataset)
        # run the transformer and get results
        activations = self.get_activations(loader)
        for section_index, section in id_section_map.items():
            words = self.section_frames_to_tokenised_words(
                section_index=section_index,
                batch_encoding=dataset.encodings,
                predictions=activations,
            )
            entities = self.tokenized_word_processor(
                words, text=section.text, namespace=self.namespace()
            )
            section.entities.extend(entities)
            if self.entity_splitter:
                for ent in entities:
                    section.entities.extend(self.entity_splitter(ent, section.text))

    def get_activations(self, loader: DataLoader) -> Tensor:
        """get a namedtuple_values_indices consisting of confidence and labels
        for a given dataloader (i.e. run bert)

        :param loader:
        :return:
        """
        results = torch.cat(
            [
                x.logits  # type: ignore[union-attr]
                for x in self.trainer.predict(  # type: ignore[union-attr]
                    model=self.model, dataloaders=loader, return_predictions=True
                )
            ]
        )
        return self.activation_fn(results)

    def get_dataloader(self, docs: list[Document]) -> tuple[DataLoader, dict[int, Section]]:
        """
        get a dataloader from a List of :class:`kazu.data.data.Document`. Collation is handled via
        :class:`transformers.DataCollatorWithPadding`\\ .

        :param docs:
        :return: The returned dict's keys map to overflow_to_sample_mapping in the
                underlying batch encoding, allowing the processing of docs longer than can fit within the maximum
                sequence length of a transformer
        """
        batch_encoding, id_section_map = documents_to_document_section_batch_encodings_map(
            docs, self.tokeniser, stride=self.stride, max_length=self.max_sequence_length
        )
        dataset = HFDataset(batch_encoding)
        collate_func = DataCollatorWithPadding(
            tokenizer=self.tokeniser, padding=PaddingStrategy.MAX_LENGTH
        )
        loader = DataLoader(dataset=dataset, batch_size=self.batch_size, collate_fn=collate_func)
        return loader, id_section_map

    def frame_to_tok_word(
        self,
        batch_encoding: BatchEncoding,
        number_of_frames: int,
        frame_index: int,
        section_frame_index: int,
        predictions: Tensor,
    ) -> list[TokenizedWord]:
        """Depending on the number of frames generated by a string of text, and
        whether it is the first or last frame, we need to return different
        subsets of the frame offsets and frame word_ids.

        :param batch_encoding:
        :param number_of_frames: number of frames created by the tokenizer for the string
        :param frame_index: the index of the query frame, relative to the total number of frames
        :param section_frame_index: the index of the section frame, relative to the whole BatchEncoding
        :param predictions:
        :return: Tuple of 2 lists: frame offsets and frame word ids
        """
        half_stride = int(self.stride / 2)
        # 1:-1 skip cls and sep tokens
        if number_of_frames == 1:
            start_index = 1
            end_index = -1
        elif number_of_frames > 1 and frame_index == 0:
            start_index = 1
            end_index = -(half_stride + 1)
        elif number_of_frames > 1 and frame_index == number_of_frames - 1:
            start_index = half_stride + 1
            end_index = -1
        else:
            start_index = half_stride + 1
            end_index = -(half_stride + 1)

        assert batch_encoding.encodings is not None
        frame_offsets = batch_encoding.encodings[section_frame_index].offsets[start_index:end_index]
        frame_word_ids: list[Optional[int]] = batch_encoding.encodings[
            section_frame_index
        ].word_ids[start_index:end_index]
        frame_token_ids = batch_encoding.encodings[section_frame_index].ids[start_index:end_index]
        frame_tokens = batch_encoding.encodings[section_frame_index].tokens[start_index:end_index]
        predictions = predictions[section_frame_index][start_index:end_index]
        prev_word_id: Optional[int] = None
        all_words = []

        word_id_index_start, offset_start, offset_end = 0, 0, 0

        for i, word_id in enumerate(frame_word_ids):
            if word_id != prev_word_id:
                # start a new word, and add the previous
                if prev_word_id is not None:
                    all_words.append(
                        TokenizedWord(
                            token_offsets=frame_offsets[word_id_index_start:i],
                            token_confidences=predictions[word_id_index_start:i],
                            token_ids=frame_token_ids[word_id_index_start:i],
                            tokens=frame_tokens[word_id_index_start:i],
                            word_char_start=offset_start,
                            word_char_end=offset_end - 1,
                            word_id=prev_word_id,
                        )
                    )
                word_id_index_start = i
                offset_start, offset_end = frame_offsets[i]
            if i == len(frame_word_ids) - 1 and word_id is not None:
                # if checking the last word in a frame, if word_id is not None, add it
                all_words.append(
                    TokenizedWord(
                        token_offsets=frame_offsets[word_id_index_start : i + 1],
                        token_confidences=predictions[word_id_index_start : i + 1],
                        token_ids=frame_token_ids[word_id_index_start : i + 1],
                        tokens=frame_tokens[word_id_index_start : i + 1],
                        word_char_start=offset_start,
                        word_char_end=offset_end,
                        word_id=word_id,
                    )
                )

            _, offset_end = frame_offsets[i]
            prev_word_id = word_id

        frame_word_ids_excluding_nones = set(frame_word_ids)
        frame_word_ids_excluding_nones.discard(None)
        assert len(set(frame_word_ids_excluding_nones)) == len(all_words)

        logger.debug(
            f"inputs this frame: {self.tokeniser.decode(frame_token_ids[:start_index])}<-IGNORED "
            f"START->{self.tokeniser.decode(frame_token_ids[start_index:end_index])}"
            f"<-END IGNORED->{self.tokeniser.decode(frame_token_ids[end_index:])}"
        )
        return all_words

    def section_frames_to_tokenised_words(
        self,
        section_index: int,
        batch_encoding: BatchEncoding,
        predictions: Tensor,
    ) -> list[TokenizedWord]:
        """

        :param section_index:
        :param batch_encoding:
        :param predictions:
        :return:
        """
        words = []

        section_frame_indices = self.get_list_of_batch_encoding_frames_for_section(
            batch_encoding, section_index
        )
        for frame_index, section_frame_index in enumerate(section_frame_indices):
            word_sub_list = self.frame_to_tok_word(
                batch_encoding=batch_encoding,
                number_of_frames=len(section_frame_indices),
                frame_index=frame_index,
                section_frame_index=section_frame_index,
                predictions=predictions,
            )
            words.extend(word_sub_list)
        return words

    @staticmethod
    def get_list_of_batch_encoding_frames_for_section(
        batch_encoding: BatchEncoding, section_index: int
    ) -> list[int]:
        """For a given dataloader with a HFDataset, return a list of frame
        indexes associated with a given section index.

        :param batch_encoding:
        :param section_index:
        :return:
        """
        section_frame_indices = [
            i
            for i, x in enumerate(batch_encoding.data["overflow_to_sample_mapping"])
            if x == section_index
        ]
        return section_frame_indices

    @staticmethod
    def id2labels_from_label_list(labels: Iterable[str]) -> dict[int, str]:
        return {idx: label for idx, label in enumerate(labels)}



import logging
from collections import defaultdict
from typing import Iterable

from kazu.data.data import CuratedTermBehaviour, CuratedTerm
from kazu.database.in_memory_db import SynonymDatabase
from kazu.ontology_preprocessing.base import OntologyParser

logger = logging.getLogger(__name__)


def filter_curations_for_ner(
    curations: Iterable[CuratedTerm], parser: OntologyParser
) -> Iterable[CuratedTerm]:
    """Filter curations to retain those that can be used for dictionary based
    NER.

    Also checks the curations are represented in the internal database.
    """
    original_terms = defaultdict(set)
    inherited_terms = defaultdict(set)
    syn_db = SynonymDatabase()
    for curation in curations:
        if curation.behaviour is CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING:
            original_terms[curation.curated_synonym].add(curation)
        elif curation.behaviour is CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM:
            inherited_terms[curation.source_term].add(curation)
        else:
            # not an ner behaviour, so move on to the next curation in the loop
            continue
    for curated_synonym, curation_set in original_terms.items():
        if len(curation_set) > 1:
            # note, this shouldn't happen, as it should be handled by the curation validation logic.
            # however, just in case this logic changes, we'll leave this here
            logger.warning("multiple curations detected for %s, %s", curated_synonym, curation_set)
            term_norm = next(iter(curation_set)).term_norm_for_linking(parser.entity_class)
            if term_norm not in syn_db.get_all(parser.name):
                logger.warning(
                    "dictionary based NER needs an database entry for %s, %s, but none exists",
                    term_norm,
                    parser.name,
                )
                continue
        yield from curation_set
        yield from inherited_terms.pop(curated_synonym, set())
    if len(inherited_terms) > 0:
        logger.warning(
            "The following inherited curations were unmatched to a source term, and will be ignored for NER: %s",
            inherited_terms.values(),
        )



import os
import dataclasses
import logging
from typing import Optional, Callable
from collections.abc import Iterable

try:
    from py4j.java_gateway import JavaGateway
    from rdkit import Chem
except ImportError as e:
    raise ImportError(
        "To use OpsinStep, you need to install py4j and rdkit.\n"
        "You can either install these yourself, or install kazu[all-steps].\n"
    ) from e

from kazu.data.data import Document, Entity, CharSpan, Mapping, StringMatchConfidence
from kazu.steps import Step, document_iterating_step


BREAKS = set(" !@#&?|\t\n\r")  # https://www.acdlabs.com/iupac/nomenclature/93/r93_45.htm

logger = logging.getLogger(__name__)


class OpsinStep(Step):
    """A Step that calls Opsin (Open Parser for Systematic IUPAC Nomenclature)
    over py4j.

    :py:class:`~.TransformersModelForTokenClassificationNerStep` often identifies
    `IUPAC chemical nomenclature strings <https://en.wikipedia.org/wiki/IUPAC_nomenclature_of_organic_chemistry>`_
    as :class:`~.Entity`\\ s with an :attr:`~.Entity.entity_class` of ``drug``, but these entities
    fail to map to any of the drug parsers as no synonym is present. This step provides an extra
    way to resolve these chemical entities.

    `Opsin <https://opsin.ch.cam.ac.uk/>`_ produces a
    `SMILES <https://en.wikipedia.org/wiki/Simplified_molecular-input_line-entry_system>`_
    from an IUPAC string and we use `rdkit <https://www.rdkit.org>`_ to convert that to a canonical
    SMILES to allow comparison between entities. This step then produces a :class:`~.Mapping` with
    the canonical SMILES string as the :attr:`~.Mapping.idx`\\ .

    Adding ``${OpsinStep}`` just after ``${MappingStep}`` in ``kazu/conf/Pipeline/default.yaml``
    will enable this step.

    .. attention::

        To use this step, you will need `py4j <https://www.py4j.org>`_ and
        `rdkit <https://www.rdkit.org>`_ installed, which is not installed as
        part of the default kazu install because this step isn't used as part
        of the default pipeline.

        You can either do:

        .. code-block:: console

            $ pip install py4j rdkit

        Or you can install required dependencies for all steps included in kazu
        with:

        .. code-block:: console

            $ pip install kazu[all-steps]

    .. note::

      The nature of this functionality is considered experimental and we may split it into two steps in the future, without
      making a major release. If you are using or are interested in using this step, please
      `open a GitHub issue <https://github.com/AstraZeneca/KAZU/issues/new>`_.

      .. raw:: html

        <details>
        <summary>Full details of possible change</summary>

      In particular, this step does two things:

      1. Adjust incorrect NER boundaries (particularly coming from the :class:`~.TransformersModelForTokenClassificationNerStep`\\ )
      2. Link drug entities consisting of IUPAC strings to a canonical SMILES

      The second of these aligns closely with the 'linking' stage in kazu, along with :class:`~.DictionaryEntityLinkingStep`\\ .
      We would ideally like to wrap the logic of 1. above into :class:`~.TransformersModelForTokenClassificationNerStep` like with the
      :class:`~.NonContiguousEntitySplitter` to fix these issues everywhere, and have 2. as a standalone linking step. However,
      this will require changes to the MappingLogic, and it may be tricky to de-couple 1 & 2 (the way this step currently does this
      depends on running Opsin to do accurately, which we would like to avoid doing twice, which may justify leaving this as a single
      step).

      .. raw:: html

        </details>

    .. testsetup::
        :skipif: kazu_model_pack_missing

        from pathlib import Path

        from hydra import initialize_config_dir, compose

        from kazu.utils.constants import HYDRA_VERSION_BASE

        # the hydra config is kept in the model pack
        cdir = Path(os.environ["KAZU_MODEL_PACK"]).joinpath("conf")

        with initialize_config_dir(version_base=HYDRA_VERSION_BASE, config_dir=str(cdir)):
            kazu_config = compose(config_name="config")

        from hydra.utils import instantiate

        from kazu.data.data import Mapping
        from kazu.steps.ner.opsin import OpsinStep

        opsin_step: OpsinStep = instantiate(kazu_config.OpsinStep)

        # function to make the doctests shorter
        def op(s: str) -> str:
            mapping: Optional[Mapping] = opsin_step.parseString(s)
            if mapping is None:
                return None
            else:
                return mapping.idx

    +------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    | Examples                                                                                                                                                                                             |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | IUPAC Input                                                                                                                 | SMILES Output                                                          |
    +=============================================================================================================================+========================================================================+
    | Bicyclo[3.2.1]octane                                                                                                        | C1CC2CCC(C1)C2                                                         |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op("Bicyclo[3.2.1]octane")                                      |
    |                                                                                                                             |    'C1CC2CCC(C1)C2'                                                    |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | 2,2'-ethylenedipyridine                                                                                                     | c1ccc(CCc2ccccn2)nc1                                                   |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op("2,2'-ethylenedipyridine")                                   |
    |                                                                                                                             |    'c1ccc(CCc2ccccn2)nc1'                                              |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | Benzo[1",2":3,4;4",5":3',4']dicyclobuta[1,2-b:1',2'-c']difuran                                                              | c1cc2c3cc4c5cocc5c4cc3c2o1                                             |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op(                                                             |
    |                                                                                                                             |    ...     "Benzo[1\\",2\\":3,4;4\\",5\\":3',4']dicyclobuta"               |
    |                                                                                                                             |    ...     "[1,2-b:1',2'-c']difuran"                                   |
    |                                                                                                                             |    ... )                                                               |
    |                                                                                                                             |    'c1cc2c3cc4c5cocc5c4cc3c2o1'                                        |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | Cyclohexanone ethyl methyl ketal                                                                                            | CCOC1(OC)CCCCC1                                                        |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op("Cyclohexanone ethyl methyl ketal")                          |
    |                                                                                                                             |    'CCOC1(OC)CCCCC1'                                                   |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | 4-[2-(2-chloro-4-fluoroanilino)-5-methylpyrimidin-4-yl]-N-[(1S)-1-(3-chlorophenyl)-2-hydroxyethyl]-1H-pyrrole-2-carboxamide | Cc1cnc(Nc2ccc(F)cc2Cl)nc1-c1c[nH]c(C(=O)N[C@H](CO)c2cccc(Cl)c2)c1      |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op(                                                             |
    |                                                                                                                             |    ...     "4-[2-(2-chloro-4-fluoroanilino)-5-methylpyrimidin"         |
    |                                                                                                                             |    ...     "-4-yl]-N-[(1S)-1-(3-chlorophenyl)-2-hydroxyethyl]"         |
    |                                                                                                                             |    ...     "-1H-pyrrole-2-carboxamide"                                 |
    |                                                                                                                             |    ... )                                                               |
    |                                                                                                                             |    'Cc1cnc(Nc2ccc(F)cc2Cl)nc1-c1c[nH]c(C(=O)N[C@H](CO)c2cccc(Cl)c2)c1' |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | 7-cyclopentyl-5-(4-methoxyphenyl)pyrrolo[2,3-d]pyrimidin-4-amine                                                            | COc1ccc(-c2cn(C3CCCC3)c3ncnc(N)c23)cc1                                 |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op(                                                             |
    |                                                                                                                             |    ...     "7-cyclopentyl-5-(4-methoxyphenyl)"                         |
    |                                                                                                                             |    ...     "pyrrolo[2,3-d]pyrimidin-4-amine"                           |
    |                                                                                                                             |    ... )                                                               |
    |                                                                                                                             |    'COc1ccc(-c2cn(C3CCCC3)c3ncnc(N)c23)cc1'                            |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | | [(3S,3aS,6R,6aS)-3-nitrooxy-2,3,3a,5,6,6a-hexahydrofuro[3,2-b]furan-6-yl] nitrate                                         | O=[N+]([O-])O[C@H]1CO[C@H]2[C@@H]1OC[C@H]2O[N+](=O)[O-]                |
    | | (see `pubchem <https://pubchem.ncbi.nlm.nih.gov/compound/6883>`_)                                                         |                                                                        |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op(                                                             |
    |                                                                                                                             |    ...     "[(3S,3aS,6R,6aS)-3-nitrooxy"                               |
    |                                                                                                                             |    ...     "-2,3,3a,5,6,6a-hexahydrofuro[3,2-b]furan-6-yl]"            |
    |                                                                                                                             |    ...     " nitrate"                                                  |
    |                                                                                                                             |    ... )                                                               |
    |                                                                                                                             |    'O=[N+]([O-])O[C@H]1CO[C@H]2[C@@H]1OC[C@H]2O[N+](=O)[O-]'           |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+
    | 1,4:3,6-dianhydro-2,5-di-O-Nitro-D-glucitol                                                                                 | | Opsin fails to parse this.                                           |
    |                                                                                                                             | | As a result, the Step will not produce a :class:`~.Mapping`\\ .       |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             | .. doctest::                                                           |
    |                                                                                                                             |    :hide:                                                              |
    |                                                                                                                             |    :skipif: kazu_model_pack_missing                                    |
    |                                                                                                                             |                                                                        |
    |                                                                                                                             |    >>> op("1,4:3,6-dianhydro-2,5-di-O-Nitro-D-glucitol") is None       |
    |                                                                                                                             |    True                                                                |
    +-----------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------+



    Paper:

    | Daniel M. Lowe, Peter T. Corbett, Peter Murray-Rust, and Robert C. Glen
    | Chemical Name to Structure: OPSIN, an Open Source Solution
    | Journal of Chemical Information and Modeling 2011 51 (3), 739-753
    | DOI: `10.1021/ci100384d <https://doi.org/10.1021/ci100384d>`_

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @article{doi:10.1021/ci100384d,
        author = {Lowe, Daniel M. and Corbett, Peter T. and Murray-Rust, Peter and Glen, Robert C.},
        title = {Chemical Name to Structure: OPSIN, an Open Source Solution},
        journal = {Journal of Chemical Information and Modeling},
        volume = {51},
        number = {3},
        pages = {739-753},
        year = {2011},
        doi = {10.1021/ci100384d},
            note ={PMID: 21384929},

        URL = {
                https://doi.org/10.1021/ci100384d

        },
        eprint = {
                https://doi.org/10.1021/ci100384d
        }
        }

    .. raw:: html

        </details>
    """

    def __init__(
        self,
        entity_class: str,
        opsin_fatjar_path: str,
        java_home: str,
        condition: Optional[Callable[[Document], bool]] = None,
    ):
        """
        :param entity_class: search entities of this class for resolvable IUPAC string
        :param opsin_fatjar_path: path to a py4j fatjar, containing OPSIN dependencies
        :param java_home: path to installed java runtime
        :param condition: Since OPSIN can be slow, we can optionally specify a callable, so that
            any documents that don't contain pre-existing drug entities are not processed
        """
        self.condition = condition
        if not os.path.exists(opsin_fatjar_path):
            raise RuntimeError(f"required jar: {opsin_fatjar_path} not found")
        self.gateway = JavaGateway.launch_gateway(
            jarpath=".",
            classpath=opsin_fatjar_path,
            die_on_exit=True,
            java_path=os.path.join(java_home, "bin", "java"),
        )
        self.opsin = self.gateway.jvm.com.astrazeneca.kazu.OpsinRunner()
        self.entity_class = entity_class

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        if self.condition and not self.condition(doc):
            # skip this document
            return

        for section in doc.sections:
            updated_mappings = {}
            for ent in section.entities:
                if ent.entity_class == self.entity_class:
                    if len(ent.mappings) == 0:
                        # entity mapping failed, e.g. no exact matches to dictionaries
                        # look up to two spaces out
                        for (match, start, end) in self.extendString(ent, section.text, spaces=2):
                            maybe_mapping = self.parseString(match)
                            if maybe_mapping:
                                opsin_entity = dataclasses.replace(
                                    ent,
                                    match=match,
                                    spans=frozenset([CharSpan(start=start, end=end)]),
                                    mappings={maybe_mapping},
                                )
                                updated_mappings[ent] = opsin_entity
                                break

            for original_entity, opsin_entity in updated_mappings.items():
                section.entities.remove(original_entity)
                section.entities.append(opsin_entity)

    @staticmethod
    def extendString(ent: Entity, section: str, spaces: int = 0) -> Iterable[tuple[str, int, int]]:
        """Extend a possible IUPAC Entity match to longer plausible strings.

        :class:`~.TransformersModelForTokenClassificationNerStep` tends to truncate IUPAC matches to
        a first hyphen. Here, we attempt to extend the match to try to capture the full IUPAC string.

        :param ent: the entity we suspect contains (some of) an IUPAC string
        :param section: the section the entity is contained in
        :param spaces: the maximum number of IUPAC spaces/breaks to 'extend' the match through
        :return: plausible IUPAC strings with section start and end positions, ordered
            from longest to shortest.
        """
        start = ent.start
        end = ent.end
        res: list[tuple[str, int, int]] = []
        while start > 0 and section[start - 1] not in BREAKS:
            start = start - 1
        while end < len(section) and (section[end] not in BREAKS or spaces > 0):
            if section[end] in BREAKS:
                spaces = spaces - 1
                res.append((section[start:end], start, end))
            end = end + 1

        last_result = (section[start:end], start, end)
        if len(res) == 0 or not res[-1] == last_result:
            # res[-1] == last_result can be true when the while loop above
            # is broken out of immediately after encountering a BREAK and appending a result.
            # this can happen if there is a BREAK immediately before the end of
            # a section, or if there are two BREAKS in a row at the point at which
            # spaces decreases from 1 to 0
            res.append(last_result)
        # reversed because we prefer the longest strings
        yield from reversed(res)

    def parseString(self, name: str) -> Optional[Mapping]:
        """Attempt to parse a potential IUPAC drug name into a
        :class:`~.Mapping`\\ .

        Call Opsin with a potential IUPAC drug name to see if it parses - Opsin is fast enough
        that we can afford to try many potential candidates.

        If Opsin parses the string successfully, we convert the resulting
        `SMILES <https://en.wikipedia.org/wiki/Simplified_molecular-input_line-entry_system>`_
        into a canonical SMILES using `rdkit <https://www.rdkit.org>`_\\ , and produce a
        :class:`~.Mapping` with the canonical SMILES as the :attr:`~.Mapping.idx`\\ .

        If Opsin parsing fails, ``None`` is returned. If the log level is set to debug,
        the error from Opsin for the parsing failure will be logged.

        :param name: the potential IUPAC drug name
        :return:
        """
        try:
            smiles = self.opsin.nameToStructure(name)
            if smiles is not None:
                smiles = Chem.CanonSmiles(smiles)
                mapping = Mapping(
                    default_label=name,
                    source="Opsin",
                    parser_name="Opsin",
                    idx=smiles,
                    string_match_strategy=self.namespace(),
                    disambiguation_strategy=None,
                    string_match_confidence=StringMatchConfidence.HIGHLY_LIKELY,
                )
                return mapping
        except Exception as e:
            reason = e.args[1].getMessage()
            logger.debug(f"Opsin parsing error: {reason}")
        return None



import logging
from collections import defaultdict

from kazu.data.data import Document, Entity
from kazu.steps import Step, document_iterating_step

logger = logging.getLogger(__name__)


class MergeOverlappingEntsStep(Step):
    """This step merges overlapping and nested entities.

    The final result should not allow any overlapped entities see
    algorithm description below
    """

    def __init__(
        self,
        ent_class_preferred_order: list[str],
        ignore_non_contiguous: bool = True,
    ):
        """The algorithm for selecting an entity span is as follows:

        1. group entities by location

           In this context, a location is a span of text represented by a start and end char index tuple. A location
           represents all contiguous and non-contiguous entities that in some way overlap, even if not directly. E.g.

           A overlaps B but not C. B overlaps C.

           entities A, B and C are all considered to be part of the same location

        2. sort entities within each location, picking the best according to the following sort logic:

           1. prefer entities with mappings
           2. prefer longest spans
           3. prefer entities as configured by ent_class_preferred_order (see param description below)
           4. prefer entities by level of confidence of entity mention
           5. If all above are equal, the preferred entity is selected on the basis of
              the entity class name (reverse alphabetically ordered). Warning: This last sort criteria is arbitrary


        :param ent_class_preferred_order: order of namespaces to prefer. Any partially overlapped entities are
            eliminated according to this ordering (first = higher priority). If an entity class is not specified, it's
            assumed to have a priority of 0 (a.k.a lowest)
        :param ignore_non_contiguous: should non-contiguous entities be excluded from the merge process?
        """
        # store as dict for lookup speed
        self.ignore_non_contiguous = ignore_non_contiguous
        self.ent_class_preferred_order = {
            namespace: i for i, namespace in enumerate(reversed(ent_class_preferred_order))
        }

    def select_preferred_entity(self, ents: set[Entity]) -> tuple[Entity, list[Entity]]:
        """
        :param ents:
        :return: tuple of Entity<preferred> ,list[Entity]<other entities at this location>
        """
        preferred_ents = sorted(
            ents,
            key=lambda x: (
                len(x.mappings) > 0,
                (x.end - x.start),
                self.ent_class_preferred_order.get(x.entity_class, 0),
                x.mention_confidence.value,
                x.entity_class,
            ),
            reverse=True,
        )
        return preferred_ents[0], preferred_ents[1:]

    def filter_ents_across_class(self, ents: dict[tuple[int, int], set[Entity]]) -> list[Entity]:
        """Choose the best entities per location.

        :param ents:
        :return:
        """
        to_keep = []
        for ents_set in ents.values():
            if len(ents_set) > 1:
                to_keep_this_location, to_drop = self.select_preferred_entity(ents_set)
                to_keep.append(to_keep_this_location)
                logger.debug("dropping ents %s", to_drop)
            else:
                to_keep.extend(ents_set)
        return to_keep

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        for section in doc.sections:
            if len(section.entities) > 0:
                ents_to_merge, non_contig_ents = [], []
                if self.ignore_non_contiguous:
                    for ent in section.entities:
                        if len(ent.spans) == 1:
                            ents_to_merge.append(ent)
                        else:
                            non_contig_ents.append(ent)
                else:
                    ents_to_merge = section.entities

                locations_overlapped = self.group_entities_by_location(ents_to_merge)
                section.entities = self.filter_ents_across_class(locations_overlapped)
                section.entities.extend(non_contig_ents)

    def group_entities_by_location(
        self, entities: list[Entity]
    ) -> dict[tuple[int, int], set[Entity]]:
        """

        :param entities:
        :return: dict of locations to set[Entity]
        """
        if len(entities) == 0:
            return {}
        locations_by_start = sorted(entities, key=lambda x: x.start)
        locations_overlapped: defaultdict[tuple[int, int], set[Entity]] = defaultdict(set)
        ents_this_group = set()
        start = locations_by_start[0].start
        end = locations_by_start[0].end
        for entity in locations_by_start:
            if start <= entity.start < end:
                # we're still in the same location
                if entity.end > end:
                    # expand the location to new end
                    end = entity.end
                ents_this_group.add(entity)
            else:
                assert (
                    entity.start >= end
                ), "Entities overlap in an undefined fashion. This should be impossible?"
                # we've gone beyond the 'current' location, 'submit' it to locations_overlapped
                locations_overlapped[(start, end)] = ents_this_group
                # set up new location
                ents_this_group = {entity}
                start = entity.start
                end = entity.end

        locations_overlapped[(start, end)] = ents_this_group
        return dict(locations_overlapped)



import abc
import itertools
from abc import ABC
from typing import Optional
from collections.abc import Iterable

from kazu.data.data import (
    Document,
    Mapping,
    EquivalentIdSet,
    SynonymTermWithMetrics,
    StringMatchConfidence,
    DisambiguationConfidence,
)
from kazu.database.in_memory_db import MetadataDatabase, Metadata
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.base import DEFAULT_LABEL
from kazu.steps.linking.post_processing.disambiguation.strategies import DisambiguationStrategy


class MappingFactory:
    """Factory class to produce mappings."""

    @staticmethod
    def create_mapping_from_id_sets(
        id_sets: set[EquivalentIdSet],
        parser_name: str,
        string_match_strategy: str,
        string_match_confidence: StringMatchConfidence,
        disambiguation_strategy: Optional[str],
        disambiguation_confidence: Optional[DisambiguationConfidence] = None,
        additional_metadata: Optional[dict] = None,
    ) -> Iterable[Mapping]:

        for id_set in id_sets:
            yield from MappingFactory.create_mapping_from_id_set(
                id_set=id_set,
                parser_name=parser_name,
                string_match_strategy=string_match_strategy,
                string_match_confidence=string_match_confidence,
                disambiguation_strategy=disambiguation_strategy,
                disambiguation_confidence=disambiguation_confidence,
                additional_metadata=additional_metadata,
            )

    @staticmethod
    def create_mapping_from_id_set(
        id_set: EquivalentIdSet,
        parser_name: str,
        string_match_strategy: str,
        string_match_confidence: StringMatchConfidence,
        disambiguation_strategy: Optional[str],
        disambiguation_confidence: Optional[DisambiguationConfidence] = None,
        additional_metadata: Optional[dict] = None,
    ) -> Iterable[Mapping]:
        for idx, source in id_set.ids_and_source:
            yield MappingFactory.create_mapping(
                parser_name=parser_name,
                source=source,
                idx=idx,
                string_match_strategy=string_match_strategy,
                string_match_confidence=string_match_confidence,
                disambiguation_strategy=disambiguation_strategy,
                disambiguation_confidence=disambiguation_confidence,
                additional_metadata=additional_metadata if additional_metadata is not None else {},
            )

    @staticmethod
    def _get_default_label_and_metadata_from_parser(
        parser_name: str, idx: str
    ) -> tuple[str, Metadata]:
        metadata = MetadataDatabase().get_by_idx(name=parser_name, idx=idx)
        default_label = metadata.pop(DEFAULT_LABEL)
        assert isinstance(default_label, str)
        return default_label, metadata

    @staticmethod
    def create_mapping(
        parser_name: str,
        source: str,
        idx: str,
        string_match_strategy: str,
        string_match_confidence: StringMatchConfidence,
        disambiguation_strategy: Optional[str] = None,
        disambiguation_confidence: Optional[DisambiguationConfidence] = None,
        additional_metadata: Optional[dict] = None,
        xref_source_parser_name: Optional[str] = None,
    ) -> Mapping:
        default_label, metadata = MappingFactory._get_default_label_and_metadata_from_parser(
            parser_name, idx
        )
        if additional_metadata:
            metadata.update(additional_metadata)
        return Mapping(
            default_label=default_label,
            idx=idx,
            source=source,
            string_match_strategy=string_match_strategy,
            string_match_confidence=string_match_confidence,
            disambiguation_strategy=disambiguation_strategy,
            disambiguation_confidence=disambiguation_confidence,
            parser_name=parser_name,
            metadata=metadata,
            xref_source_parser_name=xref_source_parser_name,
        )


class MappingStrategy(ABC):
    """A MappingStrategy is responsible for actualising instances of
    :class:`.Mapping`\\ .

    This is performed in two steps:

    1. Filter the set of :class:`.SynonymTermWithMetrics` associated with an :class:`.Entity` down
       to the most appropriate ones, (e.g. based on string similarity).

    2. If required, apply any configured :class:`.DisambiguationStrategy` to the filtered instances
       of :class:`.EquivalentIdSet`\\ .

    Selected instances of :class:`.EquivalentIdSet` are converted to :class:`.Mapping`\\ .
    """

    DISAMBIGUATION_NOT_REQUIRED = "disambiguation_not_required"

    def __init__(
        self,
        confidence: StringMatchConfidence,
        disambiguation_strategies: Optional[list[DisambiguationStrategy]] = None,
        disambiguation_essential: bool = False,
    ):
        """

        :param confidence: the level of confidence that should be assigned to this strategy. This is simply a label
            for human users, and has no bearing on the actual algorithm.
        :param disambiguation_strategies: after :meth:`filter_terms` is called, these strategies are triggered if either
            multiple instances of :class:`.SynonymTermWithMetrics` remain, and/or any of them are ambiguous.
        :param disambiguation_essential: disambiguation strategies MUST deliver a result, in order for this strategy to pass
        """

        self.disambiguation_essential = disambiguation_essential
        self.confidence = confidence
        if disambiguation_essential and (
            disambiguation_strategies is None or len(disambiguation_strategies) == 0
        ):
            raise ValueError(
                "disambiguation strategies must be provided, as disambiguation_essential=True"
            )
        self.disambiguation_strategies = disambiguation_strategies

    def prepare(self, document: Document) -> None:
        """Perform any setup that needs to run once per document.

        Care should be taken if trying to cache this step, as the Document state is
        liable to change between executions.

        :param document:
        :return:
        """
        pass

    @abc.abstractmethod
    def filter_terms(
        self,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        """
        Algorithms should override this method to return a set of the "best"
        :class:`.SynonymTermWithMetrics` for a given query string.

        Ideally, this will be a set with a single element. However, it may not be possible to
        identify a single best match. In this scenario, the id sets of multiple
        :class:`.SynonymTermWithMetrics` will be carried forward for disambiguation
        (if configured).

        :param ent_match: the raw entity string.
        :param ent_match_norm: normalised version of the entity string.
        :param document: originating Document.
        :param terms: terms to filter.
        :param parser_name: parser name associated with these terms.
        :return: a set of filtered terms
        """
        pass

    def disambiguate_if_required(
        self, filtered_terms: set[SynonymTermWithMetrics], document: Document, parser_name: str
    ) -> tuple[set[EquivalentIdSet], Optional[str], Optional[DisambiguationConfidence]]:
        """applies disambiguation strategies if configured, and either
        len(filtered_terms) > 1 or any of the filtered_terms are ambiguous. If
        ids are still ambiguous after all strategies have run, the
        disambiguation confidence will be
        :attr:`.DisambiguationConfidence.AMBIGUOUS`\\

        :param filtered_terms: terms to disambiguate
        :param document: originating Document
        :param parser_name: parser name associated with these terms
        :return:
        """

        all_id_sets = set(id_set for term in filtered_terms for id_set in term.associated_id_sets)

        if not self.disambiguation_essential and len(all_id_sets) == 1:
            # there's a single id set that isn't ambiguous, no need to disambiguate
            return all_id_sets, self.DISAMBIGUATION_NOT_REQUIRED, None
        elif not self.disambiguation_essential and (
            self.disambiguation_strategies is None or len(self.disambiguation_strategies) == 0
        ):
            return all_id_sets, None, DisambiguationConfidence.AMBIGUOUS
        else:
            assert self.disambiguation_strategies is not None
            for strategy in self.disambiguation_strategies:
                filtered_id_sets = strategy(
                    id_sets=all_id_sets, document=document, parser_name=parser_name
                )
                if len(filtered_id_sets) == 1:
                    return filtered_id_sets, strategy.__class__.__name__, strategy.confidence
            if self.disambiguation_essential:
                return set(), None, DisambiguationConfidence.AMBIGUOUS
            else:
                return all_id_sets, None, DisambiguationConfidence.AMBIGUOUS

    def __call__(
        self,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
    ) -> Iterable[Mapping]:
        """

        :param ent_match: unnormalised NER string match (i.e. :attr:`.Entity.match`)
        :param ent_match_norm: normalised NER string match (i.e. :attr:`.Entity.match_norm`)
        :param document: originating document
        :param terms: set of terms to consider. Note, terms from different parsers should not be mixed.
        :return:
        """
        parser_name = next(iter(terms)).parser_name

        filtered_terms = self.filter_terms(
            ent_match=ent_match,
            ent_match_norm=ent_match_norm,
            document=document,
            terms=terms,
            parser_name=parser_name,
        )

        (
            id_sets,
            successful_disambiguation_strategy,
            disambiguation_confidence,
        ) = self.disambiguate_if_required(filtered_terms, document, parser_name)
        yield from MappingFactory.create_mapping_from_id_sets(
            id_sets=id_sets,
            parser_name=parser_name,
            string_match_confidence=self.confidence,
            disambiguation_confidence=disambiguation_confidence,
            additional_metadata=None,
            string_match_strategy=self.__class__.__name__,
            disambiguation_strategy=successful_disambiguation_strategy,
        )


class ExactMatchMappingStrategy(MappingStrategy):
    """Returns any exact matches."""

    @staticmethod
    def filter_terms(
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        return set(term for term in terms if term.exact_match)


class SymbolMatchMappingStrategy(MappingStrategy):
    """Split both query and reference terms by whitespace.

    Select the term with the most splits as the 'query'. Check all of
    these tokens (and no more) are within the other term. Useful for
    symbol matching e.g. "MAP K8" (longest) vs "MAPK8" (shortest).
    """

    @staticmethod
    def match_symbols(s1: str, s2: str) -> bool:
        # the pattern should either be in both or neither
        reference_term_tokens = s1.split(" ")
        query_term_tokens = s2.split(" ")
        if len(reference_term_tokens) > len(query_term_tokens):
            longest = reference_term_tokens
            shortest = s2
        else:
            longest = query_term_tokens
            shortest = s1

        for tok in longest:
            if tok not in shortest:
                return False
            else:
                shortest = shortest.replace(tok, "", 1)

        return shortest.strip() == ""

    @classmethod
    def filter_terms(
        cls,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        return set(term for term in terms if cls.match_symbols(ent_match_norm, term.term_norm))


class TermNormIsSubStringMappingStrategy(MappingStrategy):
    """for a set of :class:`.SynonymTermWithMetrics`, see if any of their
    .term_norm are string matches of the match_norm tokens based on whitespace
    tokenisation.

    If exactly one :class:`.SynonymTermWithMetrics` matches, prefer it.

    Works best on symbolic entities, e.g. "TESTIN gene" ->"TESTIN".
    """

    def __init__(
        self,
        confidence: StringMatchConfidence,
        disambiguation_strategies: Optional[list[DisambiguationStrategy]] = None,
        disambiguation_essential: bool = False,
        min_term_norm_len_to_consider: int = 3,
    ):
        """

        :param confidence:
        :param disambiguation_strategies:
        :param disambiguation_essential:
        :param min_term_norm_len_to_consider: only consider instances of
            :class:`.SynonymTermWithMetrics` where the length of :attr:`~.SynonymTerm.term_norm` is
            equal to or greater than this value.
        """
        super().__init__(
            confidence, disambiguation_strategies, disambiguation_essential=disambiguation_essential
        )
        self.min_term_norm_len_to_consider = min_term_norm_len_to_consider

    def filter_terms(
        self,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        norm_tokens = set(ent_match_norm.split(" "))

        filtered_terms_and_len = [
            (
                term,
                len(term.term_norm),
            )
            for term in terms
            if term.term_norm in norm_tokens
            and len(term.term_norm) >= self.min_term_norm_len_to_consider
        ]
        filtered_terms_and_len.sort(key=lambda x: x[1], reverse=True)
        filtered_terms_by_len_groups = itertools.groupby(filtered_terms_and_len, key=lambda x: x[1])
        for _, terms_and_len_iter in filtered_terms_by_len_groups:
            terms_and_len_list = list(terms_and_len_iter)
            if len(terms_and_len_list) == 1:
                return {terms_and_len_list[0][0]}
        return set()


class StrongMatchMappingStrategy(MappingStrategy):
    """
    1. sort :class:`.SynonymTermWithMetrics` by highest scoring search match to identify the
       highest scoring match.
    2. query remaining matches to see whether their scores are greater than this best score - the
       differential (i.e. there are many close string matches).
    """

    def __init__(
        self,
        confidence: StringMatchConfidence,
        disambiguation_strategies: Optional[list[DisambiguationStrategy]] = None,
        disambiguation_essential: bool = False,
        search_threshold: float = 80.0,
        symbolic_only: bool = False,
        differential: float = 2.0,
    ):
        """

        :param confidence:
        :param disambiguation_strategies:
        :param disambiguation_essential:
        :param search_threshold: only consider synonym terms above this search threshold
        :param symbolic_only: only consider terms that are symbolic
        :param differential: only consider terms with search scores equal or greater to the best match minus this value
        """
        super().__init__(
            confidence, disambiguation_strategies, disambiguation_essential=disambiguation_essential
        )
        self.differential = differential
        self.symbolic_only = symbolic_only
        self.search_threshold = search_threshold

    def filter_terms(
        self,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        if self.symbolic_only:
            relevant_terms_with_scores = [
                (term, score)
                for term in terms
                if term.is_symbolic and (score := term.search_score) is not None
            ]
        else:
            relevant_terms_with_scores = [
                (term, score) for term in terms if (score := term.search_score) is not None
            ]

        if len(relevant_terms_with_scores) == 0:
            return set()

        best_score = max((score for _, score in relevant_terms_with_scores))

        return set(
            term
            for term, score in relevant_terms_with_scores
            if score >= self.search_threshold and best_score - score <= self.differential
        )


class StrongMatchWithEmbeddingConfirmationStringMatchingStrategy(StrongMatchMappingStrategy):
    """
    Same as parent class, but a complex string scorer with a predefined threshold is used to confirm that the
    ent_match is broadly similar to one of the terms attached to the
    :class:`.SynonymTermWithMetrics`\\ .

    Useful for refining non-symbolic close string matches (e.g. "Neck disease" and "Heck disease").
    """

    def __init__(
        self,
        confidence: StringMatchConfidence,
        complex_string_scorer: StringSimilarityScorer,
        disambiguation_strategies: Optional[list[DisambiguationStrategy]] = None,
        disambiguation_essential: bool = False,
        search_threshold: float = 80.0,
        embedding_threshold: float = 0.60,
        symbolic_only: bool = False,
        differential: float = 2.0,
    ):
        """
        :param confidence:
        :param complex_string_scorer: only consider synonym terms passing this string scorer call
        :param disambiguation_strategies:
        :param disambiguation_essential:
        :param search_threshold: only consider synonym terms above this search threshold
        :param embedding_threshold: the Entity.match and one of the SynonymTermWithMetrics.terms must be
            above this threshold (according to the complex_string_scorer) for the term to be valid
        :param symbolic_only: only consider terms that are symbolic
        :param differential: only consider terms with search scores equal or greater to the best match minus this value
        """
        super().__init__(
            confidence=confidence,
            search_threshold=search_threshold,
            differential=differential,
            symbolic_only=symbolic_only,
            disambiguation_strategies=disambiguation_strategies,
            disambiguation_essential=disambiguation_essential,
        )
        self.embedding_threshold = embedding_threshold
        self.complex_string_scorer = complex_string_scorer

    def filter_terms(
        self,
        ent_match: str,
        ent_match_norm: str,
        document: Document,
        terms: frozenset[SynonymTermWithMetrics],
        parser_name: str,
    ) -> set[SynonymTermWithMetrics]:
        synonym_term_sorted_by_score = sorted(
            super().filter_terms(
                ent_match=ent_match,
                ent_match_norm=ent_match_norm,
                document=document,
                terms=terms,
                parser_name=parser_name,
            ),
            key=lambda x: x.search_score,  # type: ignore[arg-type,return-value]
            reverse=True,
        )
        selected_id_sets = set()
        selected_terms = set()
        for term in synonym_term_sorted_by_score:
            if term.associated_id_sets not in selected_id_sets:
                selected_id_sets.add(term.associated_id_sets)
                if any(
                    self.complex_string_scorer(ent_match, original_term) >= self.embedding_threshold
                    for original_term in term.terms
                ):
                    selected_terms.add(term)
        return selected_terms



import logging
import os
import sys
import tempfile
from pathlib import Path
from typing import Any, TypeVar, Protocol, Optional, Union
from collections.abc import Iterable, Callable

from cachetools import LFUCache
from diskcache import Cache

from kazu.data.data import Entity, SynonymTermWithMetrics
from kazu.utils.utils import get_match_entity_class_hash

logger = logging.getLogger(__name__)
kazu_model_pack_dir = os.getenv("KAZU_MODEL_PACK")
KAZU_DISK_CACHE_NAME = "kazu_disk_cache"


Ret = TypeVar("Ret", covariant=True)
"""A TypeVar to represent a return value.

Used in :class:`~.Memoization` to encode that a memorized function
returns the same type as the original function it 'memoized'.
"""


class Memoization(Protocol[Ret]):
    def __cache_key__(self, *args: Any, **kwargs: Any) -> tuple[Any, ...]:
        raise NotImplementedError

    def __call__(self, *args: Any, **kwargs: Any) -> Ret:
        raise NotImplementedError


class CacheProtocol(Protocol):
    def memoize(
        self,
        name: Optional[str] = None,
        typed: bool = False,
        expire: Optional[float] = None,
        tag: Optional[str] = None,
        ignore: set[Union[str, int]] = set(),
    ) -> Callable[[Callable[..., Ret]], Memoization[Ret]]:
        raise NotImplementedError

    def clear(self) -> int:
        raise NotImplementedError

    def delete(self, key: Any) -> bool:
        raise NotImplementedError


kazu_disk_cache: CacheProtocol
"""We use the :class:`diskcache.Cache` concept to cache expensive to produce
resources to disk. Methods and functions can be decorated with
kazu_disk_cache.memoize() to use this feature. Note, when used with a class
method, the default behaviour of this function is to generate a key based on
the constructor arguments of the class instance. Since these can be large (e.g.
OntologyParser), we sometimes use the ignore argument to override this
behaviour.

e.g.

.. code-block:: python

    @kazu_disk_cache.memoize(ignore={0})
    def method_of_class_with_lots_of_args(self):
        ...
"""
if kazu_model_pack_dir is None:
    kazu_disk_cache_path_str = tempfile.mkdtemp(suffix=KAZU_DISK_CACHE_NAME)
    logger.warning(
        "KAZU_MODEL_PACK env variable not set. Using %s for caching. "
        "This will not be reused if the process is exited",
        kazu_disk_cache_path_str,
    )
else:
    kazu_disk_cache_path_str = str(Path(kazu_model_pack_dir).joinpath(KAZU_DISK_CACHE_NAME))
    logger.info("using disk cache at %s", kazu_disk_cache_path_str)

kazu_disk_cache = Cache(
    directory=kazu_disk_cache_path_str,
    eviction_policy="none",
    size_limit=sys.maxsize,
    cull_limit=0,
)


class EntityLinkingLookupCache:
    """A simple wrapper around LFUCache to reduce calls to expensive processes
    (e.g. bert)"""

    def __init__(self, lookup_cache_size: int = 5000):
        self.terms_lookup_cache: LFUCache[int, set[SynonymTermWithMetrics]] = LFUCache(
            lookup_cache_size
        )

    def update_terms_lookup_cache(
        self, entity: Entity, terms: Iterable[SynonymTermWithMetrics]
    ) -> None:
        hash_val = get_match_entity_class_hash(entity)
        cache_hit = self.terms_lookup_cache.get(hash_val)
        if cache_hit is None:
            self.terms_lookup_cache[hash_val] = set(terms)

    def check_lookup_cache(self, entities: Iterable[Entity]) -> list[Entity]:
        """checks the cache for synonym terms. If relevant terms are found for
        an entity, update it accordingly. If not return as a list of cache
        misses (e.g. for further processing)

        :param entities:
        :return:
        """
        cache_misses = []
        for ent in entities:
            hash_val = get_match_entity_class_hash(ent)
            terms_from_cache = self.terms_lookup_cache.get(hash_val, set())
            if not terms_from_cache:
                cache_misses.append(ent)
            else:
                ent.update_terms(terms_from_cache)
        return cache_misses



import dataclasses
import functools
import itertools
import json
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from copy import deepcopy
from typing import Optional

from tqdm import tqdm
from spacy.matcher import Matcher

from kazu.data.data import SynonymTerm
from kazu.language.language_phenomena import GREEK_SUBS, DASHES
from kazu.utils.spacy_pipeline import SpacyPipelines, BASIC_PIPELINE_NAME, basic_spacy_pipeline
from kazu.utils.utils import PathLike

logger = logging.getLogger(__name__)


class SynonymGenerator(ABC):
    @abstractmethod
    def call(self, synonym_str: str) -> Optional[set[str]]:
        """Implementations should override this method to generate new strings
        from an input string."""
        pass

    @functools.cache
    def generate_synonyms(self, synonym_str: str) -> Optional[set[str]]:
        """Cached wrapper around :meth:`.SynonymGenerator.call`\\.

        Caching prevents recomputation of redundant strings.
        """
        return self.call(synonym_str)

    def __call__(self, synonym_terms: set[SynonymTerm]) -> set[SynonymTerm]:

        existing_terms = set(term for synonym in synonym_terms for term in synonym.terms)

        result: set[SynonymTerm] = set()
        for synonym_term in tqdm(
            synonym_terms,
            total=len(synonym_terms),
            desc=f"generating synonyms for {self.__class__.__name__}",
        ):
            for synonym_str in synonym_term.terms:
                generated_synonym_strings = self.generate_synonyms(synonym_str)
                if generated_synonym_strings:
                    new_terms: set[str] = set()
                    for generated_syn in generated_synonym_strings:
                        if generated_syn in existing_terms:
                            logger.debug(
                                "generated synonym %s matches existing synonym %s",
                                generated_syn,
                                synonym_term,
                            )
                        elif generated_syn in new_terms:
                            logger.debug(
                                "generated synonym %s matches another generated synonym %s",
                                generated_syn,
                                synonym_term,
                            )
                        else:
                            new_terms.add(generated_syn)
                    if new_terms:
                        if synonym_term.original_term is not None:
                            orig_term = synonym_term.original_term
                        else:
                            orig_term = synonym_str
                        result.add(
                            dataclasses.replace(
                                synonym_term, original_term=orig_term, terms=frozenset(new_terms)
                            )
                        )
        return result


class CombinatorialSynonymGenerator:
    def __init__(self, synonym_generators: Iterable[SynonymGenerator]):
        self.synonym_generators: set[SynonymGenerator] = set(synonym_generators)

    def __call__(self, synonyms: set[SynonymTerm]) -> set[SynonymTerm]:
        """For every permutation of modifiers, generate a list of syns, then
        aggregate at the end.

        :param synonyms: input to generation process
        :return: generated synonyms. Note, the field values of the generated synonyms will
            be the same as the seed synonym, apart from SynonymTerm.terms, which contains the generated synonyms
        """
        synonym_gen_permutations = list(itertools.permutations(self.synonym_generators))
        results = set()
        logger.info(
            "Running synonym generation permutations. This may be slow at first, but will speed up as caching takes effect."
        )
        for i, permutation_list in enumerate(synonym_gen_permutations):
            # make a copy of the original synonyms
            all_syns = deepcopy(synonyms)
            logger.info(
                "running permutation set %s of %s. Permutations: %s",
                i + 1,
                len(synonym_gen_permutations),
                permutation_list,
            )
            for generator in permutation_list:
                # run the generator
                new_syns: set[SynonymTerm] = generator(all_syns)
                results.update(new_syns)
                all_syns.update(new_syns)

        for generator in self.synonym_generators:
            generator.generate_synonyms.cache_clear()
        return results


# TODO: this isn't used currently - do we want to try and refine it
# or just kill it off altogether?
class SeparatorExpansion(SynonymGenerator):
    def __init__(self):
        self.all_stopwords = basic_spacy_pipeline().Defaults.stop_words
        self.end_expression_brackets = r"(.*)\((.*)\)$"
        self.mid_expression_brackets = r"(.*)\(.*\)(.*)"
        self.excluded_parenthesis = ["", "non-protein coding"]

    def call(self, synonym_str: str) -> Optional[set[str]]:
        bracket_results = set()
        all_group_results = set()
        if "(" in synonym_str and ")" in synonym_str:
            # expand end expression brackets
            matches = re.match(self.end_expression_brackets, synonym_str)
            if matches is not None:
                all_groups_no_brackets = []
                for group in matches.groups():
                    if (
                        group not in self.excluded_parenthesis
                        and group.lower() not in self.all_stopwords
                    ):
                        bracket_results.add(group.strip())
                        all_groups_no_brackets.append(group)
                all_group_results.add("".join(all_groups_no_brackets))
            else:
                # remove mid expression  brackets
                matches = re.match(self.mid_expression_brackets, synonym_str)
                if matches is not None:
                    all_groups_no_brackets = []
                    for group in matches.groups():
                        all_groups_no_brackets.append(group.strip())
                    all_group_results.add(" ".join(all_groups_no_brackets))

        # expand slashes
        for x in list(bracket_results):
            if "/" in x:
                splits = x.split("/")
                for split in splits:
                    bracket_results.add(split.strip())
            if "," in x:
                splits = x.split(",")
                for split in splits:
                    bracket_results.add(split.strip())
        bracket_results.update(all_group_results)
        if len(bracket_results) > 0:
            return bracket_results
        else:
            return None


class StopWordRemover(SynonymGenerator):
    """Remove stopwords from a string."""

    all_stopwords = {"of", "and", "in", "to", "with", "caused", "involved", "by", "the"}

    @classmethod
    def call(cls, synonym_str: str) -> Optional[set[str]]:
        new_terms = set()
        lst = []
        detected = False
        for token in synonym_str.split():
            if token.lower() in cls.all_stopwords:
                detected = True
            else:
                lst.append(token)
        if detected:
            new_terms.add(" ".join(lst))
        if new_terms:
            return new_terms
        else:
            return None


class GreekSymbolSubstitution:

    ALL_SUBS: dict[str, set[str]] = defaultdict(set)
    for greek_letter, spelling in GREEK_SUBS.items():

        ALL_SUBS[greek_letter].add(spelling)
        # single letter abbreviation
        ALL_SUBS[greek_letter].add(spelling[0])
        # reversed
        ALL_SUBS[spelling].add(greek_letter)

        # replace between lower and upper case
        # to generate synonyms
        if greek_letter.islower():
            upper_greek_letter = greek_letter.upper()
            ALL_SUBS[greek_letter].add(upper_greek_letter)
        elif greek_letter.isupper():
            lower_greek_letter = greek_letter.lower()
            ALL_SUBS[greek_letter].add(lower_greek_letter)

    # we don't want this to have the semantics of a defaultdict for missing lookups
    ALL_SUBS = dict(ALL_SUBS)


class StringReplacement(SynonymGenerator):
    GREEK_VARIANT_PREFIX_SUFFIX = DASHES.union(set(" "))

    def __init__(
        self,
        replacement_dict: Optional[dict[str, list[str]]] = None,
        digit_aware_replacement_dict: Optional[dict[str, list[str]]] = None,
        include_greek: bool = True,
    ):
        self.include_greek = include_greek
        self.replacement_dict = replacement_dict
        self.digit_aware_replacement_dict = digit_aware_replacement_dict

    def call(self, synonym_str: str) -> Optional[set[str]]:
        results = set()
        if self.replacement_dict:
            for to_replace, replacement_list in self.replacement_dict.items():
                if to_replace in synonym_str:
                    for replace_with in replacement_list:
                        results.add(synonym_str.replace(to_replace, replace_with).strip())
        if self.digit_aware_replacement_dict:
            for to_replace, replacement_list in self.digit_aware_replacement_dict.items():
                matches = set(re.findall(to_replace + r"[0-9]+", synonym_str))
                for match in matches:
                    number = match.split(to_replace)[1]
                    for sub_in in replacement_list:
                        new_str = synonym_str.replace(match, f"{sub_in}{number}").strip()
                        results.add(new_str)

        if self.include_greek:
            self._generate_greek_subs(results, synonym_str)

        if len(results) > 0:
            return results
        else:
            return None

    def _generate_greek_subs(self, results, synonym_str):
        # only strip text once initially - the greek character replacement
        # will not introduce leading or trailing whitespace unlike the other
        # replacements above
        stripped_text = synonym_str.strip()
        strings_to_substitute = {stripped_text}
        for candidate, replacement_set in GreekSymbolSubstitution.ALL_SUBS.items():
            # if it's in the original text it should be in all previous substitutions, no
            # need to check all of them
            for fix in self.GREEK_VARIANT_PREFIX_SUFFIX:
                prefix = False
                suffix = False
                # necessary so we don't modify strings_to_substitute while looping over it,
                # which throws an error
                outputs_this_step = set()
                if f"{fix}{candidate}" in synonym_str:
                    suffix = True
                if f"{candidate}{fix}" in synonym_str:
                    prefix = True

                for string_to_substitute in strings_to_substitute:
                    for replacement in replacement_set:
                        if prefix:
                            single_unique_letter_substituted = string_to_substitute.replace(
                                f"{candidate}{fix}", f"{replacement}{fix}"
                            )
                            outputs_this_step.add(single_unique_letter_substituted)
                            results.add(single_unique_letter_substituted)
                        if suffix:
                            single_unique_letter_substituted = string_to_substitute.replace(
                                f"{fix}{candidate}", f"{fix}{replacement}"
                            )
                            outputs_this_step.add(single_unique_letter_substituted)
                            results.add(single_unique_letter_substituted)
                strings_to_substitute.update(outputs_this_step)


class SuffixReplacement(SynonymGenerator):
    """Interchange all suffixes within a provided set to produce new synonyms.

    Note, this is expected to be noisy, and for most of the generated synonyms not to be valid
    words. This class is present as a generation step for high recall, with curation of synonyms
    expected later.

    In particular, note that this also doesn't check for the longest matching suffix - e.g. for a
    synonym 'anaemia' and the suffixes 'ia', 'a' and 'ic', the new synonyms 'anaemic' and
    'amaemiic' will both be generated.
    """

    def __init__(self, suffixes: Iterable[str]):
        self.suffixes = set(suffixes)

    def call(self, synonym_str: str) -> Optional[set[str]]:
        new_terms: set[str] = set()
        for suffix in self.suffixes:
            # Note that this will trigger twice for 'ia' since 'a' is also present.
            # We expect this to be noisy, and then curate from this.
            if synonym_str.endswith(suffix):
                term_without_suffix = synonym_str.removesuffix(suffix)
                new_terms.update(
                    term_without_suffix + new_suffix
                    for new_suffix in self.suffixes
                    if new_suffix is not suffix
                )

        if new_terms:
            return new_terms
        else:
            return None


class SpellingVariationReplacement(SynonymGenerator):
    """Generate additional synonyms using a mapping of (known) synonyms to a
    list of variations."""

    def __init__(self, input_path: PathLike):
        with open(input_path, mode="r", encoding="utf-8") as inf:
            raw_variation_mapping: dict[str, list[str]] = json.load(inf)

        # lowercase for case-insensitive comparison
        self.variation_mapping = {k.lower(): val for k, val in raw_variation_mapping.items()}

    def call(self, synonym_str: str) -> Optional[set[str]]:
        new_terms = set()
        variations = self.variation_mapping.get(synonym_str.lower())
        if variations is not None:
            new_terms.update(variations)
        if new_terms:
            return new_terms
        else:
            return None


class NgramHyphenation(SynonymGenerator):
    """Generate hyphenated variants of ngrams."""

    def __init__(self, ngram: int = 2):
        self.ngram = ngram

    def call(self, synonym_str: str) -> Optional[set[str]]:
        parts = synonym_str.split()
        if len(parts) != self.ngram:
            return None
        else:
            new_terms = set()
            for hyphen in DASHES:
                new_terms.add(hyphen.join(parts))
        return new_terms


class TokenListReplacementGenerator(SynonymGenerator):
    """Given lists of tokens, generate an alternative string based upon a query
    token.

    Note, this implementation is pretty basic, and only replaces one
    token at a time. It's mainly designed for ontologies like Meddra
    which stretch the definition of an entity somewhat, by incorporating
    verbs (e.g. "increase in AST").
    """

    def __init__(
        self,
        token_lists_to_consider: list[list[str]],
    ):
        """

        :param token_lists_to_consider: if any token from the sublist matches a query string, generate
            new strings based upon all tokens in this sublist.
        """

        SpacyPipelines().add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)
        SpacyPipelines().add_reload_callback_func(BASIC_PIPELINE_NAME, self._init_token_matcher)
        self.token_lists_to_consider = token_lists_to_consider
        self._init_token_matcher()

    def _init_token_matcher(self) -> None:
        matcher = Matcher(SpacyPipelines.get_model(BASIC_PIPELINE_NAME).vocab)
        for i, token_list in enumerate(self.token_lists_to_consider):
            matcher.add(key=i, patterns=[[{"LOWER": {"IN": token_list}}]])
        self.token_matcher = matcher

    def call(self, synonym_str: str) -> Optional[set[str]]:
        new_terms = set()
        doc = SpacyPipelines().process_single(text=synonym_str, model_name=BASIC_PIPELINE_NAME)
        matches = self.token_matcher(doc)
        if matches is not None:
            for match_id, match_start, match_end in matches:
                found_tokens = doc[match_start:match_end].text
                variant_list = self.token_lists_to_consider[match_id]
                for variant in variant_list:
                    new_terms.add(synonym_str.replace(found_tokens, variant))
        return new_terms


class VerbPhraseVariantGenerator(SynonymGenerator):
    """Generate alternative verb phrases based on a list of tense templates,
    and lemmas matched in a query.

    It's mainly designed for ontologies like Meddra which stretch the
    definition of an entity somewhat, by incorporating verbs (e.g.
    "increase in AST").
    """

    NOUN_PLACEHOLDER = "NOUN"
    VERB_PLACEHOLDER = "TARGET"

    def __init__(
        self,
        tense_templates: list[str],
        lemmas_to_consider: dict[str, list[str]],
        spacy_model_path: str,
    ):
        """

        :param tense_templates: template expressons to generate, for example:

            .. code-block:: python

                ["{NOUN} {TARGET}", "{TARGET} in {NOUN}"]

        :param lemmas_to_consider: a dict of verb lemmas to surface forms to generate, for example:

            .. code-block:: python

                {"increase": ["increasing", "increased"], "decrease": ["decreased", "decreasing"]}

        :param spacy_model_path: path to a serialised spacy model - must have a lemmatizer component.
        """
        SpacyPipelines().add_from_path(spacy_model_path, spacy_model_path)
        SpacyPipelines().add_reload_callback_func(spacy_model_path, self._init_lemma_matcher)
        self.spacy_model_path = spacy_model_path
        self.tense_templates = tense_templates
        self.lemmas_to_consider = lemmas_to_consider
        self._init_lemma_matcher()

    def _init_lemma_matcher(self) -> None:
        matcher = Matcher(SpacyPipelines.get_model(self.spacy_model_path).vocab)
        matcher.add(key=0, patterns=[[{"LEMMA": {"IN": list(self.lemmas_to_consider.keys())}}]])
        self.lemma_matcher = matcher

    def _populate_lemma_template(
        self, template: str, lemma: str, surface_forms: list[str], noun: str
    ) -> Iterable[str]:
        for form in [lemma] + surface_forms:
            yield template.format(**{self.NOUN_PLACEHOLDER: noun, self.VERB_PLACEHOLDER: form})

    def call(self, synonym_str: str) -> Optional[set[str]]:
        new_terms: set[str] = set()
        doc = SpacyPipelines().process_single(text=synonym_str, model_name=self.spacy_model_path)
        noun_matches = self.lemma_matcher(doc)
        if noun_matches is not None:
            for _match_id, match_start, _ in noun_matches:
                verb_lemma = None
                noun = []
                for i, tok in enumerate(doc):
                    if i == match_start:
                        verb_lemma = tok.lemma_
                    else:
                        noun.append(tok.text)
                if len(noun) > 0 and verb_lemma is not None:
                    noun_str = " ".join(noun)
                    surface_forms = self.lemmas_to_consider[verb_lemma]
                    for template in self.tense_templates:
                        new_terms.update(
                            self._populate_lemma_template(
                                template=template,
                                surface_forms=surface_forms,
                                noun=noun_str,
                                lemma=verb_lemma,
                            )
                        )
        return new_terms



from itertools import groupby
from typing import TypeVar, TYPE_CHECKING
from collections.abc import Iterable, Callable

# see https://github.com/python/typeshed/tree/master/stdlib/_typeshed
# In short, this is needed as
# the _typeshed package and its types do not exist at runtime
if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison

Item = TypeVar("Item")
"""The type of an element in the iterable being grouped."""
# we need to use Supports RichComparison here otherwise mypy complains
# about the 'key' argument to sorted
Key = TypeVar("Key", bound="SupportsRichComparison")
"""The type of the sort key provided by the key_func.

Bound to 'SupportsRichComparison' from
`_typeshed <https://github.com/python/typeshed/tree/main/stdlib/_typeshed>`_
as the keys must support comparison in order to be sorted using :func:`sorted`\\ .
"""


def sort_then_group(
    items: Iterable[Item], key_func: Callable[[Item], Key]
) -> Iterable[tuple[Key, Iterable[Item]]]:
    """.. without the override below, it fails to find Item and Key

    ..
      This is due to an issue with Sphinx handling builtins - see
      https://github.com/sphinx-doc/sphinx/issues/11571
      as it used to work while using typing.Callable and typing.Iterable
      instead of the collections.abc versions. Switching to show JsonDictType
      instead as it's simpler to write this override, plus I think more
      readable for users.
      An alternative would be to attempt to use Sphinx's autodoc_type_aliases
      https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_type_aliases
      but this requires doing ``from __future__ import annotations`` which could
      break pydantic stuff and have wider codebase implications, so this would
      be a potentially larger piece of work for not much gain.

    :param items:
    :type items: ~collections.abc.Iterable[~kazu.utils.grouping.Item]
    :param key_func:
    :type key_func: ~collections.abc.Callable[[~kazu.utils.grouping.Item], ~kazu.utils.grouping.Key]
    :rtype: ~collections.abc.Iterable[tuple[~kazu.utils.grouping.Key], ~collections.abc.Iterable[~kazu.utils.grouping.Item]]
    """
    sorted_items = sorted(items, key=key_func)
    yield from groupby(sorted_items, key=key_func)



KAZU = "kazu"
NO_AUTH_ENDPOINTS = [
    "/api",
    "/api/",
    "/api/docs",
    "/api/openapi.json",
    "/api/kazu/ls-annotations",
]
NO_AUTH_DIRS = [
    # nothing under /ui requires auth
    # the main page is /ui/index, but
    # it needs to be able to load static resources
    # which are places under /ui as well.
    "/ui"
]



from typing import Optional

from kazu.data.data import Document
from kazu.steps import Step, document_iterating_step
from kazu.utils.abbreviation_detector import KazuAbbreviationDetector


class AbbreviationFinderStep(Step):
    """Detects abbreviations using the algorithm in "A simple algorithm for
    identifying abbreviation definitions in biomedical text.", (Schwartz &
    Hearst, 2003).

    Uses a modified version of the scispacy abbreviation finder rules, to expand abbreviations (see
    :class:`kazu.utils.abbreviation_detector.KazuAbbreviationDetector`\\ ). In this implementation,
    abbreviations learnt in one section will be applied throughout the others.
    """

    def __init__(self, exclude_abbrvs: Optional[list[str]] = None):
        self.detector = KazuAbbreviationDetector(
            namespace=self.namespace(), exclude_abbrvs=exclude_abbrvs
        )

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        """

        :param doc:
        :return:
        """
        self.detector(doc)



import logging
from collections.abc import Iterable

from kazu.data.data import Document
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.steps import document_iterating_step
from kazu.steps.linking.post_processing.strategy_runner import StrategyRunner
from kazu.steps import ParserDependentStep

logger = logging.getLogger(__name__)


class MappingStep(ParserDependentStep):
    """A wrapper for :class:`.StrategyRunner`\\, so it can be used in a
    pipeline."""

    def __init__(
        self,
        parsers: Iterable[OntologyParser],
        strategy_runner: StrategyRunner,
    ):
        super().__init__(parsers)
        self.strategy_runner = strategy_runner

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        self.strategy_runner(doc)



import spacy

from kazu.ontology_matching.ontology_matcher import OntologyMatcher, SPAN_KEY
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.utils.utils import PathLike
from kazu.utils.spacy_pipeline import basic_spacy_pipeline


def main(
    output_dir: PathLike, parsers: list[OntologyParser], span_key: str = SPAN_KEY
) -> spacy.language.Language:
    """Generates, serializes and returns a Spacy pipeline with an
    :class:`~kazu.ontology_matching.ontology_matcher.OntologyMatcher`.

    Generates an English Spacy pipeline with a tokenizer, a sentencizer with default
    config, and an OntologyMatcher based on the input parameters. The pipeline is
    written to disk, and also returned to the caller.

    If a parser has no curations configured, the
    :class:`~kazu.ontology_matching.ontology_matcher.OntologyMatcher` is built directly
    from the configured synonyms from the parsers (with any associated generated synonyms). This is
    useful for trying to understand which strings are 'noisy', but not recommended for production as raw ontology
    data tends to need some curation before it can be applied.

    If a parser has curations configured, it will be built using the configured instances of :class:`~kazu.data.data.CuratedTerm`
    associated with the parser. This is generally the recommended behaviour in production, although obviously
    requires a set of high quality curations to be effective

    :param output_dir: the output directory to write the pipeline into.
    :param parsers: build the pipeline using these parsers as a data source.
    :param span_key: the key to use within the generated Spacy Docs'
        `span attribute <https://spacy.io/api/doc#spans>`_ to store and access recognised NER
        spans.
    """
    nlp = basic_spacy_pipeline()
    config = {
        "span_key": span_key,
        "parser_name_to_entity_type": {parser.name: parser.entity_class for parser in parsers},
    }
    ontology_matcher = nlp.add_pipe("ontology_matcher", config=config)
    assert isinstance(ontology_matcher, OntologyMatcher)
    ontology_matcher.create_phrasematchers_using_curations(parsers)
    nlp.to_disk(output_dir)
    return nlp



import logging
from collections.abc import Iterator, Iterable
from typing import cast

from kazu.data.data import (
    CharSpan,
    Document,
    Section,
    Entity,
    SynonymTermWithMetrics,
    MentionConfidence,
)
from kazu.database.in_memory_db import SynonymDatabase
from kazu.ontology_matching import assemble_pipeline
from kazu.ontology_matching.ontology_matcher import OntologyMatcher, _MatcherOntologyData
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.steps import document_batch_step
from kazu.steps import ParserDependentStep
from kazu.utils.spacy_pipeline import SpacyPipelines
from kazu.utils.utils import PathLike, as_path
from spacy.tokens import Span, Doc

logger = logging.getLogger(__name__)


class ExplosionStringMatchingStep(ParserDependentStep):
    """A wrapper for the explosion ontology-based entity matcher and linker."""

    def __init__(
        self,
        parsers: Iterable[OntologyParser],
        path: PathLike,
        include_sentence_offsets: bool = True,
        ignore_cache: bool = False,
    ):
        """


        :param path: path to spacy pipeline including Ontology Matcher.
        :param include_sentence_offsets: whether to add sentence offsets to the metadata.
        :param ignore_cache: ignore cached version of spacy pipeline (if available) and rebuild

        """
        super().__init__(parsers)
        self.include_sentence_offsets = include_sentence_offsets
        self.path = as_path(path)

        if self.path.exists() and not ignore_cache:
            logger.info("loading spacy pipeline from %s", str(path))
            SpacyPipelines.add_from_path(name=self.namespace(), path=str(self.path.absolute()))
        else:
            logger.info(
                "cached spacy pipeline not detected or ignore_cache=True. Creating it at %s. This may take some time",
                str(self.path),
            )
            assemble_pipeline.main(output_dir=self.path, parsers=list(parsers))
            SpacyPipelines.add_from_path(name=self.namespace(), path=str(self.path.absolute()))

        matcher = cast(
            OntologyMatcher, SpacyPipelines.get_model(self.namespace()).get_pipe("ontology_matcher")
        )
        self.span_key = matcher.span_key

        self.synonym_db = SynonymDatabase()
        self.spacy_pipelines = SpacyPipelines()

    def extract_entity_data_from_spans(
        self, spans: Iterable[Span]
    ) -> Iterator[tuple[int, int, str, _MatcherOntologyData]]:
        for span in spans:
            yield span.start_char, span.end_char, span.text, span._.ontology_dict_

    @document_batch_step
    def __call__(self, docs: list[Document]) -> None:
        texts_and_sections = ((section.text, section) for doc in docs for section in doc.sections)

        # TODO: multiprocessing within the pipe command?
        spacy_result = cast(
            Iterable[tuple[Doc, Section]],
            self.spacy_pipelines.process_batch(
                texts=texts_and_sections, model_name=self.namespace(), as_tuples=True
            ),
        )

        for processed_text, section in spacy_result:
            entities = []

            spans = processed_text.spans[self.span_key]
            for (
                start_char,
                end_char,
                text,
                ontology_data,
            ) in self.extract_entity_data_from_spans(spans):

                for entity_class, per_parser_term_norm_set in ontology_data.items():
                    confidences = set()
                    e = Entity.load_contiguous_entity(
                        start=start_char,
                        end=end_char,
                        match=text,
                        entity_class=entity_class,
                        namespace=self.namespace(),
                    )
                    terms = []
                    for parser_name, term_norm, confidence in per_parser_term_norm_set:
                        mention_confidence = MentionConfidence(int(confidence))
                        confidences.add(mention_confidence)
                        term_with_metrics = SynonymTermWithMetrics.from_synonym_term(
                            self.synonym_db.get(parser_name, term_norm), exact_match=True
                        )
                        terms.append(term_with_metrics)
                    e.update_terms(terms)
                    e.mention_confidence = max(confidences)
                    entities.append(e)

            # add sentence offsets
            if self.include_sentence_offsets:
                section.sentence_spans = (
                    CharSpan(sent.start_char, sent.end_char) for sent in processed_text.sents
                )

            # if one section of a doc fails after others have succeeded, this will leave failed docs
            # in a partially processed state. It's actually unclear to me whether this is desireable or not.
            section.entities.extend(entities)



from torch import nn

from transformers import BertModel, BertPreTrainedModel


# ignore required because transformers doesn't distribute type information
# so to mypy, this is subclassing 'Any'.
class TinyBertForSequenceTagging(BertPreTrainedModel):  # type: ignore[misc]
    """PyTorch BERT model for sequence tagging.

    Based off `TinyBERT from Huawei Noah's Ark Lab <https://github.com/huawei-noah/Pretrained-Language-Model/tree/master/TinyBERT>`_
    - the `TinyBertForSequenceClassification <https://github.com/huawei-noah/Pretrained-Language-Model/blob/master/TinyBERT/transformer/modeling.py#L1119>`_
    class specifically.

    Modified for distillation using Pytorch Lightning by KAZU team.

    Originally Licensed under Apache 2.0

    | Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team., and KAZU Team
    | Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
    | Copyright 2020 Huawei Technologies Co., Ltd

    .. raw:: html

        <details>
        <summary>Full License Notice</summary>

    | Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team., and KAZU Team
    | Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
    | Copyright 2020 Huawei Technologies Co., Ltd

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    .. raw:: html

        </details>
    """

    def __init__(self, config, num_labels=None, fit_size=768):
        super(TinyBertForSequenceTagging, self).__init__(config)

        if num_labels is None:
            self.num_labels = config.num_labels
        else:
            self.num_labels = num_labels
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(in_features=config.hidden_size, out_features=self.num_labels)
        self.fit_dense = nn.Linear(config.hidden_size, fit_size)
        self.init_weights()

    def forward(
        self, input_ids, token_type_ids=None, attention_mask=None, labels=None, is_student=False
    ):
        """Defines the computation performed when the model is called.

        Note that users should call the
        :class:`TinyBertForSequenceTagging` instance itself, rather than
        this method directly, because calling the instance runs
        registered 'hooks' on the instance.

        This works as this class inherits (through its base class) from
        :class:`torch.nn.Module`\\ , which defines __call__ to call the
        forward method, as well as registered hooks.
        """

        output = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_hidden_states=True,
            output_attentions=True,
            return_dict=True,
        )
        sequence_output = output["hidden_states"]
        # sequence_output [list of torch tensor] = (number of layers + 1) * [batch_size, sequence_length, hidden_size]
        att_output = output["attentions"]
        logits = self.classifier(sequence_output[-1])

        tmp = []
        if is_student:
            for sequence_layer in sequence_output:
                tmp.append(self.fit_dense(sequence_layer))
            sequence_output = tmp
        return logits, att_output, sequence_output



import functools
import logging
import os
from collections import defaultdict
from collections.abc import Callable, Iterable
from copy import deepcopy
from string import ascii_lowercase
from typing import Union, Any, overload, Literal

import spacy
from kazu.utils.utils import Singleton
from spacy.lang.char_classes import (
    LIST_ELLIPSES,
    LIST_ICONS,
    CONCAT_QUOTES,
    ALPHA_LOWER,
    ALPHA_UPPER,
    ALPHA,
    HYPHENS,
)
from spacy.lang.en import English, EnglishDefaults
from spacy.lang.en.tokenizer_exceptions import TOKENIZER_EXCEPTIONS
from spacy.language import Language
from spacy.tokens import Doc

logger = logging.getLogger(__name__)

SPACY_DEFAULT_INFIXES = (
    LIST_ELLIPSES
    + LIST_ICONS
    + [
        r"(?<=[0-9])[+\-\*^](?=[0-9-])",
        r"(?<=[{al}{q}])\.(?=[{au}{q}])".format(al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        # note: this will get removed below
        r"(?<=[{a}0-9])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),
        r"(?<=[{a}0-9])[:<>=/](?=[{a}])".format(a=ALPHA),
    ]
)

# We remove the hypen infix pattern because splitting by hyphens causes problems like
# splitting the company name 'ALK-Abello' into ['ALK', '-', 'Abello']
# and then potentially recognising ALK as a gene.

# A spacy PR that modified this hyphen pattern specifically mentions biomedical texts
# as being potentially problematic for this rule:
# https://github.com/explosion/spaCy/pull/5770#issuecomment-659389160
DEFAULT_INFIXES_MINUS_HYPHENS = SPACY_DEFAULT_INFIXES[:-2] + SPACY_DEFAULT_INFIXES[-1:]

# There are exceptions saying not to tokenize things like 'a.' and 'b.' - presumably
# to handle lists elements with letters like this. That isn't relevant for how we're
# using spacy to do NER + linking, and it breaks some cases for us - when an entity
# like 'Haemophilia A' is at the end of a sentence, so we get a token like 'A.', which
# then doesn't get split, so we don't recognise it, unless we remove these exceptions.
TOKENIZER_EXCEPTIONS_MINUS_SINGLE_LETTER = deepcopy(TOKENIZER_EXCEPTIONS)
for letter in ascii_lowercase:
    TOKENIZER_EXCEPTIONS_MINUS_SINGLE_LETTER.pop(letter + ".")


class KazuCustomEnglishDefaults(EnglishDefaults):
    tokenizer_exceptions = TOKENIZER_EXCEPTIONS_MINUS_SINGLE_LETTER
    infixes = [r"\(", "/"] + DEFAULT_INFIXES_MINUS_HYPHENS
    # related to the above, '.' isn't picked up as a suffix when preceded by a
    # single uppercase character
    suffixes = EnglishDefaults.suffixes + [  # type:ignore[operator] # because mypy
        r"(?<=\b[{au}])\.".format(
            au=ALPHA_UPPER
        )  # doesn't know that EnglishDefaults.suffixes is always a list[str]
    ]


@spacy.registry.languages("kazu_custom_en")
class KazuCustomEnglish(English):
    lang = "kazu_custom_en"
    Defaults = KazuCustomEnglishDefaults


BASIC_PIPELINE_NAME = "basic"


def basic_spacy_pipeline() -> Language:
    """A basic Spacy pipeline with a sentence splitter and a customised
    tokenizer."""
    nlp = spacy.blank("kazu_custom_en")
    nlp.add_pipe("sentencizer")
    return nlp


class SpacyPipelines(metaclass=Singleton):
    """Wraps spacy pipelines into a singleton, so multiple can be accessed from
    different locations without additional memory overhead.

    In addition, due to a
    `known memory issue <https://github.com/explosion/spaCy/discussions/9362>`_,
    we reload each pipeline after a certain number of calls, namely :attr:`~.reload_at`.

    .. note::
        In order for the Garbage Collector to remove old spacy Vocab objects, users
        should ensure that `spacy.Doc <https://spacy.io/api/doc>`_ objects are de-referenced as soon as possible
        (i.e. that you don't keep the results of :meth:`~.process_single` or :meth:`~.process_batch` around for
        a long time).
    """

    def __init__(self):
        self.reload_at = int(os.getenv("KAZU_SPACY_RELOAD_INTERVAL", 1000))
        """The interval (number of calls) at which spacy models are reloaded.

        Normally set within the kazu model pack config using the environment variable
        ``KAZU_SPACY_RELOAD_INTERVAL``, but if this isn't set (either using the config
        or as a normal environment variable), this defaults to 1000.

        .. note::
            As this class is a singleton, modifying this will change the reload value for all
            spacy pipelines (i.e. globally).

        :type: int
        """
        self.name_to_func: dict[str, Callable[[], Language]] = {}
        self.name_to_reload_callbacks: defaultdict[str, list[Callable[[], None]]] = defaultdict(
            list
        )
        self.call_counter: defaultdict[str, int] = defaultdict(lambda: 0)
        self.name_to_model: dict[str, Language] = {}

    @staticmethod
    def add_from_path(name: str, path: str) -> None:
        """Add a spacy model from a path.

        Convenience function to call :meth:`~.add_from_func` with a
        wrapped version of
        `spacy.load <https://spacy.io/api/top-level#spacy.load>`_
        with the relevant path argument.
        """
        SpacyPipelines.add_from_func(name=name, func=functools.partial(spacy.load, path))

    @staticmethod
    def add_from_func(name: str, func: Callable[[], Language]) -> None:
        """Add a spacy model from a callable."""

        instance = SpacyPipelines()
        if name in instance.name_to_func:
            logger.info("The spacy pipeline key %s is already loaded.", name)
        else:
            instance.name_to_func[name] = func
            instance.name_to_model[name] = func()

    @staticmethod
    def add_reload_callback_func(name: str, func: Callable[[], None]) -> None:
        """Add a callback when a model is reloaded.

        If using spacy components outside the context of a
        `Language <https://spacy.io/api/language>`_, these will also need to be reloaded when the
        underlying model is reloaded. This can be done by providing a zero
        argument, None return type callable. If you need to modify a field of an object with a
        callback, we recommend the use of :func:`functools.partial` to build the callback function.

        :param name:
        :param func:
        :return:
        """
        instance = SpacyPipelines()
        instance.name_to_reload_callbacks[name].append(func)

    @staticmethod
    def get_model(name: str) -> Language:
        """Get the underlying `Language <https://spacy.io/api/language>`_ for a
        given model key."""
        return SpacyPipelines().name_to_model[name]

    @overload
    def process_batch(
        self,
        texts: Iterable[tuple[Union[str, Doc], Any]],
        model_name: str,
        as_tuples: Literal[True],
        **kwargs: Any,
    ) -> Iterable[tuple[Doc, Any]]:
        ...

    @overload
    def process_batch(
        self,
        texts: Iterable[Union[str, Doc]],
        model_name: str,
        as_tuples: Literal[False] = False,
        **kwargs: Any,
    ) -> Iterable[Doc]:
        ...

    def process_batch(
        self,
        texts,
        model_name,
        as_tuples=False,
        **kwargs,
    ):
        """Process an iterable of `spacy.Doc`_ or strings with a given spacy
        model.

        :param texts: either an iterable of 'texts'
            (`spacy.Doc`_\\ s or :class:`str`\\ s)
            if ``as_tuples=False`` (the default),
            or an iterable of length 2 tuples ``(text, context)`` where context
            is an arbitrary python object that is provided back in the output
            alongside its matching text.
        :param model_name: spacy model to process texts with
        :param as_tuples: If set to ``True`` (the default is ``False``), texts are paired with
            a 'context' inside a tuple both in input and output. See the description of the
            texts argument and the output, and read the type information.
        :param kwargs: passed to `Language.pipe <https://spacy.io/api/language#pipe>`_\\ .
        :return: Iterable of `spacy.Doc`_\\ s if ``as_tuples=False``
            (the default).
            If ``as_tuples=True``, the result is an Iterable of ``(doc, context)`` tuples.

        .. _spacy.Doc: https://spacy.io/api/doc
        """
        for result in self.name_to_model[model_name].pipe(
            texts=texts, as_tuples=as_tuples, **kwargs
        ):
            yield result
            self.call_counter[model_name] += 1

        self._reload_if_required(model_name)

    def process_single(self, text: Union[str, Doc], model_name: str, **kwargs: Any) -> Doc:
        """Process a single `Doc <https://spacy.io/api/doc>`_ or text with a
        given spacy model.

        :param text:
        :param model_name:
        :param kwargs: passed to the call method of `Language <https://spacy.io/api/language>`_
        :return:
        """
        res = self.name_to_model[model_name](text, **kwargs)
        self.call_counter[model_name] += 1
        self._reload_if_required(model_name)
        return res

    def _reload_if_required(
        self,
        model_name: str,
    ) -> None:
        if self.call_counter[model_name] >= self.reload_at:
            logger.info(
                "max spacy calls (%s) exceeded for %s, "
                "see https://github.com/explosion/spaCy/discussions/9362 for more info",
                self.reload_at,
                model_name,
            )
            self.reload_model(model_name)
            self.call_counter[model_name] = 0

    def reload_model(self, model_name: str) -> None:
        """Reload a model, clearing the spacy vocab."""
        func = self.name_to_func[model_name]
        logger.info("The model will be reloaded from %s.", func)
        self.name_to_model[model_name] = func()
        for callback in self.name_to_reload_callbacks[model_name]:
            callback()



import re
from os import getenv

import regex
from functools import lru_cache
from typing import Optional, Protocol

from kazu.language.language_phenomena import GREEK_SUBS, DASHES


class EntityClassNormalizer(Protocol):
    """Protocol describing methods a normalizer should implement."""

    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        """method to determine whether a string is a symbol (e.g. "AD") or a
        noun phrase (e.g. "Alzheimers Disease")

        :param original_string:
        :return:
        """
        pass

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        """Method for normalising a symbol.

        :param original_string:
        :return:
        """
        pass

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        """Method for normalising a noun phrase.

        :param original_string:
        :return:
        """
        pass


class DefaultStringNormalizer(EntityClassNormalizer):
    """normalize a biomedical string for search.

    Suitable for most use cases
    """

    allowed_additional_chars = {" ", "(", ")", "+", "-", "‐"}
    greek_subs = GREEK_SUBS
    greek_subs_upper = {x: f" {y.upper()} " for x, y in greek_subs.items()}
    other_subs = {
        "(": " (",
        ")": ") ",
        ",": " ",
        "/": " ",
        "VIII": " 8 ",
        "VII": " 7 ",
        "XII": " 12 ",
        "III": " 3 ",
        "VI": " 6 ",
        "IV": " 4 ",
        "IX": " 9 ",
        "XI": " 11 ",
        "II": " 2 ",
    }
    re_subs = {
        re.compile(r"(?<!\()-(?!\))"): " ",  # minus not in brackets
        re.compile(r"(?<!\()‐(?!\))"): " ",  # hyphen not in brackets
        re.compile(r"\sI\s|\sI$"): " 1 ",
        re.compile(r"\sV\s|\sV$"): " 5 ",
        re.compile(r"\sX\s|\sX$"): " 10 ",
    }
    re_subs_2 = {
        re.compile(r"\sA\s|\sA$|^A\s"): " ALPHA ",
        re.compile(r"\sB\s|\sB$|^B\s"): " BETA ",
    }

    number_split_pattern = re.compile(r"(\d+)")

    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        """checks for ratio of upper to lower case characters, and numeric to
        alpha characters.

        :param original_string:
        :return:
        """
        upper_count = 0
        lower_count = 0
        numeric_count = 0

        tokens = original_string.split(" ")
        token_count = len(tokens)

        if token_count == 1 and len(original_string) <= 3:
            return True

        for i, char in enumerate(original_string):
            if char.isalpha():
                if char.isupper():
                    upper_count += 1
                    if i > 0 and token_count == 1:
                        # if is single token, and any char apart from first is upper, assume symbol
                        return True

                else:
                    lower_count += 1

            elif char.isnumeric():
                if token_count == 1:
                    # if is single token and has a number in it, assume symbol
                    return True
                numeric_count += 1

        if upper_count >= lower_count:
            return True
        elif numeric_count >= (upper_count + lower_count):
            return True
        else:
            return False

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        return " ".join(original_string.upper().split())

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        string = DefaultStringNormalizer.replace_substrings(original_string)
        string = DefaultStringNormalizer.split_on_numbers(string)
        string = DefaultStringNormalizer.replace_greek(string)
        string = DefaultStringNormalizer.remove_non_alphanum(string)
        string = DefaultStringNormalizer.depluralize(string)
        string = DefaultStringNormalizer.sub_greek_char_abbreviations(string)
        return " ".join(string.upper().split())

    @staticmethod
    def depluralize(string):
        """Apply some depluralisation rules.

        :param string:
        :return:
        """
        if len(string) > 3:
            string = GildaUtils.depluralize(string)[0]
        return string

    @staticmethod
    def sub_greek_char_abbreviations(string):
        """
        substitute single characters for alphanumeric representation - e.g. A -> ALPHA B--> BETA

        :param string:
        :return:
        """
        for re_sub, replace in DefaultStringNormalizer.re_subs_2.items():
            string = re_sub.sub(replace, string)
        return string

    @staticmethod
    def handle_lower_case_prefixes(string):
        """Preserve case only if first char of contiguous subsequence is lower
        case, and is alphanum, and upper case detected in rest of part.
        Currently unused as it causes problems with normalisation of e.g.
        erbB2, which is a commonly used form of the symbol.

        :param string:
        :return:
        """
        parts = string.split(" ")
        new_parts = []
        for part in parts:
            if part != "":
                if part.islower() and not len(part) == 1:
                    new_parts.append(part.upper())
                else:
                    first_char_case = part[0].islower()
                    if (first_char_case and part[0].isalnum()) or (
                        first_char_case and len(part) == 1
                    ):
                        new_parts.append(part)
                    else:
                        new_parts.append(part.upper())
        string = " ".join(new_parts)
        return string

    @staticmethod
    def remove_non_alphanum(string):
        """Removes all non alphanumeric characters.

        :param string:
        :return:
        """
        return "".join(
            x
            for x in string
            if (x.isalnum() or x in DefaultStringNormalizer.allowed_additional_chars)
        )

    @staticmethod
    def replace_greek(string):
        """Replaces greek characters with string representation.

        :param string:
        :return:
        """
        for substr, replace in DefaultStringNormalizer.greek_subs_upper.items():
            if substr in string:
                string = string.replace(substr, replace)
        return string

    @staticmethod
    def split_on_numbers(string):
        """Splits a string on numbers, for consistency.

        :param string:
        :return:
        """
        return " ".join(
            x.strip() for x in DefaultStringNormalizer.number_split_pattern.split(string)
        )

    @staticmethod
    def replace_substrings(original_string):
        """Replaces a range of other strings that might be confusing to a
        classifier, such as roman numerals.

        :param original_string:
        :return:
        """
        string = original_string
        for substr, replace in DefaultStringNormalizer.other_subs.items():
            if substr in string:
                string = string.replace(substr, replace)
        for re_sub, replace in DefaultStringNormalizer.re_subs.items():
            string = re_sub.sub(replace, string)
        return string


class DiseaseStringNormalizer(EntityClassNormalizer):
    known_disease_short_nouns = {"flu", "Flu", "HIV", "STI", "NSCLC"}

    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        if original_string in DiseaseStringNormalizer.known_disease_short_nouns:
            return False
        else:
            return DefaultStringNormalizer.is_symbol_like(original_string)

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        """Revert to DefaultStringNormalizer.normalize_symbol.

        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_symbol(original_string)

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        """
        revert to DefaultStringNormalizer.normalize_noun_phrase
        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_noun_phrase(original_string)


class AnatomyStringNormalizer(EntityClassNormalizer):
    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        # anatomy tends not to have symbolic representations
        return False

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        """revert to DefaultStringNormalizer.normalize_noun_phrase (note, since
        all anatomy is non-symbolic, this is theoretically superfluous, but we
        include it anyway)

        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_noun_phrase(original_string)

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        """
        revert to DefaultStringNormalizer.normalize_noun_phrase
        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_noun_phrase(original_string)


class GeneStringNormalizer(EntityClassNormalizer):
    gene_name_suffixes = {"in", "ase", "an", "gen", "gon"}

    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        """a symbol classifier that is designed to improve recall on natural
        text, especially gene symbols looks at the ratio of upper case to lower
        case chars, and the ratio of integer to alpha chars. If the ratio of
        upper case or integers is higher, assume it's a symbol. Also if the
        first char is lower case, and any subsequent characters are upper case,
        it's probably a symbol (e.g. erbB2)

        :param original_string:
        :return:
        """
        string = GildaUtils.replace_dashes(original_string, " ")
        tokens = string.split(" ")
        if len(tokens) == 1 and not any(
            tokens[0].endswith(suffix) for suffix in GeneStringNormalizer.gene_name_suffixes
        ):
            # single tokens are likely gene symbols, or can be easily classified as gene symbols - can't trust
            # capitalisation e.g. mTOR, erbBB2, egfr vs EGFR, Insulin
            # in part because of convention to camel case animal homologous gene symbols
            return True
        else:
            return all(len(x) < 4 or GeneStringNormalizer.gene_token_classifier(x) for x in tokens)

    @staticmethod
    def gene_token_classifier(original_string):
        """slightly modified version of DefaultStringNormalizer.is_symbol_like,
        designed to work on single tokens. Checks if the casing of the symbol
        changes from lower to upper (if so, is likely to be symbolic, e.g.
        erbB2)

        :param original_string:
        :return:
        """
        upper_count = 0
        lower_count = 0
        numeric_count = 0
        first_char_is_lower = len(original_string) > 0 and original_string[0].islower()
        for char in original_string:
            if char.isalpha():
                if char.isupper():
                    upper_count += 1
                    if first_char_is_lower:
                        return True
                else:
                    lower_count += 1
            elif char.isnumeric():
                numeric_count += 1
        if upper_count >= lower_count:
            return True
        elif numeric_count >= (upper_count + lower_count):
            return True
        else:
            return False

    @staticmethod
    def remove_trailing_s_if_otherwise_capitalised(string: str) -> str:
        """Frustratingly, some gene symbols are pluralised like ERBBs. we can't
        just remove trailing s as this breaks genuine symbols like 'MDH-s' and
        'GASP10ps'. So, we only strip the trailing 's' if the char before is
        upper case.

        :param string:
        :return:
        """
        if len(string) >= 3 and string[-2].isupper():
            return string.removesuffix("s")
        else:
            return string

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        """Contrary to other entity classes, gene symbols require special
        handling because of their highly unusual nature.

        :param original_string:
        :return:
        """
        string = GeneStringNormalizer.remove_trailing_s_if_otherwise_capitalised(original_string)
        string = DefaultStringNormalizer.replace_substrings(string)
        string = DefaultStringNormalizer.split_on_numbers(string)
        string = DefaultStringNormalizer.replace_greek(string)
        string = DefaultStringNormalizer.remove_non_alphanum(string)
        string = DefaultStringNormalizer.sub_greek_char_abbreviations(string)
        return " ".join(string.upper().split())

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        """Revert to DefaultStringNormalizer.normalize_noun_phrase for non
        symbolic genes.

        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_noun_phrase(original_string)


class CompanyStringNormalizer(EntityClassNormalizer):
    @staticmethod
    def is_symbol_like(original_string: str) -> bool:
        alpha_chars = filter(lambda x: x.isalpha(), original_string)
        return all(x.isupper() for x in alpha_chars)

    @staticmethod
    def normalize_symbol(original_string: str) -> str:
        """just upper case.

        :param original_string:
        :return:
        """
        return original_string.upper()

    @staticmethod
    def normalize_noun_phrase(original_string: str) -> str:
        """
        revert to DefaultStringNormalizer.normalize_noun_phrase
        :param original_string:
        :return:
        """
        return DefaultStringNormalizer.normalize_noun_phrase(original_string)


class StringNormalizer:
    """Call custom entity class normalizers, or a default normalizer if none is
    available."""

    normalizers: dict[Optional[str], type[EntityClassNormalizer]] = {
        "gene": GeneStringNormalizer,
        "anatomy": AnatomyStringNormalizer,
        "disease": DiseaseStringNormalizer,
        "company": CompanyStringNormalizer,
    }

    @staticmethod
    @lru_cache(maxsize=int(getenv("KAZU_STRING_NORMALIZER_CACHE_SIZE", 5000)))
    def classify_symbolic(original_string: str, entity_class: Optional[str] = None) -> bool:
        return StringNormalizer.normalizers.get(
            entity_class, DefaultStringNormalizer
        ).is_symbol_like(original_string)

    @staticmethod
    @lru_cache(maxsize=int(getenv("KAZU_STRING_NORMALIZER_CACHE_SIZE", 5000)))
    def normalize(original_string: str, entity_class: Optional[str] = None) -> str:
        normaliser_for_entity_class = StringNormalizer.normalizers.get(
            entity_class, DefaultStringNormalizer
        )
        if normaliser_for_entity_class.is_symbol_like(original_string):
            return normaliser_for_entity_class.normalize_symbol(original_string)
        else:
            return normaliser_for_entity_class.normalize_noun_phrase(original_string)


class GildaUtils:
    """Functions derived from `gilda <https://github.com/indralab/gilda>`_ used
    by (some of) the :py:class:`~StringNormalizer`\\ s.

    Original Credit:

    | https://github.com/indralab/gilda
    | https://github.com/indralab/gilda/blob/9e383213098144fe82103a3a5aa1bf4c14059e57/gilda/process.py

    Paper:

    | Benjamin M Gyori, Charles Tapley Hoyt, and Albert Steppi. 2022.
    | `Gilda: biomedical entity text normalization with machine-learned disambiguation as a service. <https://doi.org/10.1093/bioadv/vbac034>`_
    | Bioinformatics Advances. Vbac034.

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @article{gyori2022gilda,
            author = {Gyori, Benjamin M and Hoyt, Charles Tapley and Steppi, Albert},
            title = "{{Gilda: biomedical entity text normalization with machine-learned disambiguation as a service}}",
            journal = {Bioinformatics Advances},
            year = {2022},
            month = {05},
            issn = {2635-0041},
            doi = {10.1093/bioadv/vbac034},
            url = {https://doi.org/10.1093/bioadv/vbac034},
            note = {vbac034}
        }

    .. raw:: html

        </details>


    Licensed under BSD 2-Clause.

    Copyright (c) 2019, Benjamin M. Gyori, Harvard Medical School All rights reserved.

    .. raw:: html

        <details>
        <summary>Full License</summary>

    BSD 2-Clause License

    Copyright (c) 2019, Benjamin M. Gyori, Harvard Medical School
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this
      list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    .. raw:: html

        </details>
    """

    PLURAL_CAPS_S_PATTERN: regex.Pattern = regex.compile(r"^\p{Lu}+$")

    @staticmethod
    def depluralize(word: str) -> tuple[str, str]:
        """Return the depluralized version of the word, along with a status
        flag.

        :param word: The word which is to be depluralized.
        :return: The first element is the original word, if it is detected to be non-plural, or the
            depluralized version of the word.

            The second element is a status flag representing the detected pluralization status of
            the word, with non_plural (e.g., BRAF), plural_oes (e.g., mosquitoes),
            plural_ies (e.g., antibodies), plural_es (e.g., switches),
            plural_cap_s (e.g., MAPKs), and plural_s (e.g., receptors).
        """
        # If the word doesn't end in s, we assume it's not plural
        if not word.endswith("s"):
            return word, "non_plural"
        # Another case is words ending in -sis (e.g., apoptosis), these are almost
        # exclusively non plural so we return here too
        elif word.endswith("sis"):
            return word, "non_plural"
        # This is the case when the word ends with an o which is pluralized as oes
        # e.g., mosquitoes
        elif word.endswith("oes"):
            return word[:-2], "plural_oes"
        # This is the case when the word ends with a y which is pluralized as ies,
        # e.g., antibodies
        elif word.endswith("ies"):
            return word[:-3] + "y", "plural_ies"
        # These are the cases where words form plurals by adding -es so we
        # return by stripping it off
        elif word.endswith(("xes", "ses", "ches", "shes")):
            return word[:-2], "plural_es"
        # If the word is all caps and the last letter is an s, then it's a very
        # strong signal that it is pluralized so we have a custom return value
        # for that
        elif GildaUtils.PLURAL_CAPS_S_PATTERN.match(word[:-1]):
            return word[:-1], "plural_caps_s"
        # Otherwise, we just go with the assumption that the last s is the
        # plural marker
        else:
            return word[:-1], "plural_s"
        # Note: there don't seem to be any compelling examples of -f or -fe -> ves
        # so it is not implemented

    @classmethod
    def replace_dashes(cls, s: str, rep: str = "-") -> str:
        """Replace all types of dashes in a given string with a given
        replacement.

        :param s: The string in which all types of dashes should be replaced.
        :param rep: The string with which dashes should be replaced. By default, the plain
            ASCII dash (-) is used.
        :return: The string in which dashes have been replaced.
        """
        for d in DASHES:
            s = s.replace(d, rep)
        return s
